import importlib.util
import os
from dataclasses import dataclass
from typing import Optional

import click
from jinja2 import Environment, FileSystemLoader, Template
from linkml_runtime.linkml_model.meta import TypeDefinition

from linkml._version import __version__
from linkml.generators.oocodegen import OOCodeGenerator
from linkml.utils.generator import shared_arguments

default_template = """
{#-

  Jinja2 Template for a Java class with Lombok @Data annotation
  Annotation details at https://projectlombok.org
-#}
package {{ doc.package }};

import java.util.List;
import lombok.*;


{% if metamodel_version %}/* metamodel_version: {{metamodel_version}} */{% endif %}
{% if model_version %}/* version: {{model_version}} */{% endif %}


{% if cls.source_class.description %}/**
  {{ cls.source_class.description }}
**/{% endif %}
@Data
@EqualsAndHashCode(callSuper=false)
public{% if cls.abstract %} abstract{% endif %} class {{ cls.name }} {% if cls.is_a -%} extends {{ cls.is_a }} {%- endif %} {
{% for f in cls.fields %}
  private {{f.range}} {{ f.name }};
{%- endfor %}

}"""  # noqa: E501

TYPEMAP = {
    "xsd:string": "String",
    "xsd:integer": "Integer",
    "xsd:float": "Float",
    "xsd:double": "Double",
    "xsd:boolean": "boolean",
    "xsd:dateTime": "ZonedDateTime",
    "xsd:date": "LocalDate",
    "xsd:time": "Instant",
    "xsd:anyURI": "String",
    "xsd:decimal": "BigDecimal",
}

TYPE_DEFAULTS = {"boolean": "false", "int": "0", "float": "0f", "double": "0d", "String": '""'}


@dataclass
class JavaGenerator(OOCodeGenerator):
    """
    Generates java code from a LinkML schema.

    Two styles are supported:

    - java classes, using lombok annotations
    - java records
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["java"]
    file_extension = "java"

    # ObjectVars
    generate_records: bool = False
    """If True then use java records (introduced in java 14) rather than classes"""

    template_file: Optional[str] = None

    gen_classvars: bool = True
    gen_slots: bool = True
    genmeta: bool = False
    emit_metadata: bool = True

    def default_value_for_type(self, typ: str) -> str:
        return TYPE_DEFAULTS.get(typ, "null")

    def map_type(self, t: TypeDefinition, required: bool = False) -> str:
        if t.uri:
            # only return a Integer, Double Float when required == false
            typ = TYPEMAP.get(t.uri)
            if required and (typ == "Double" or typ == "Float"):
                typ = typ.lower()
            elif required and typ == "Integer":
                typ = "int"
            return typ
        elif t.typeof:
            return self.map_type(self.schemaview.get_type(t.typeof))
        else:
            raise ValueError(f"{t} cannot be mapped to a type")

    def serialize(self, directory: str, **kwargs) -> None:
        if self.generate_records:
            package_dir = os.path.dirname(importlib.util.find_spec(__name__).origin)
            javagen_folder = os.path.join(package_dir, "javagen", "")
            loader = FileSystemLoader(javagen_folder)
            env = Environment(loader=loader)
            template_obj = env.get_template("java_record_template.jinja2")
        elif self.template_file is not None:
            with open(self.template_file) as template_file:
                template_obj = Template(template_file.read())
        else:
            template_obj = Template(default_template)

        oodocs = self.create_documents()
        self.directory = directory
        for oodoc in oodocs:
            cls = oodoc.classes[0]
            code = template_obj.render(
                doc=oodoc,
                cls=cls,
                metamodel_version=self.schema.metamodel_version,
                model_version=self.schema.version,
            )

            os.makedirs(directory, exist_ok=True)
            filename = f"{oodoc.name}.java"
            path = os.path.join(directory, filename)
            with open(path, "w", encoding="UTF-8") as stream:
                stream.write(code)


@shared_arguments(JavaGenerator)
@click.option(
    "--output-directory",
    default="output",
    show_default=True,
    help="Output directory for individually generated class files",
)
@click.option("--package", help="Package name where relevant for generated class files")
@click.option("--template-file", help="Optional jinja2 template to use for class generation")
@click.option(
    "--generate-records/--no-generate-records",
    default=False,
    help="Optional Java 17 record implementation",
)
@click.version_option(__version__, "-V", "--version")
@click.command(name="java")
def cli(
    yamlfile,
    output_directory=None,
    package=None,
    template_file=None,
    generate_records=False,
    head=True,
    emit_metadata=False,
    genmeta=False,
    classvars=True,
    slots=True,
    **args,
):
    """Generate java classes to represent a LinkML model"""
    JavaGenerator(
        yamlfile,
        package=package,
        template_file=template_file,
        generate_records=generate_records,
        emit_metadata=head,
        genmeta=genmeta,
        gen_classvars=classvars,
        gen_slots=slots,
        **args,
    ).serialize(output_directory, **args)


if __name__ == "__main__":
    cli()



import os
import sys

from linkml_runtime.linkml_model import linkml_files
from linkml_runtime.linkml_model.linkml_files import Format, Source
from rdflib.plugins.serializers.turtle import TurtleSerializer

assert sys.version_info > (
    3,
    7,
    0,
), f"LinkML requires python 3.7.1 or later to run.  Current version: {sys.version_info}"

MODULE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

# Local location of yaml files
LOCAL_METAMODEL_YAML_FILE = linkml_files.LOCAL_PATH_FOR(Source.META, Format.YAML)
LOCAL_TYPES_YAML_FILE = linkml_files.LOCAL_PATH_FOR(Source.TYPES, Format.YAML)
LOCAL_MAPPINGS_YAML_FILE = linkml_files.LOCAL_PATH_FOR(Source.MAPPINGS, Format.YAML)
LOCAL_ANNOTATIONS_YAML_FILE = linkml_files.LOCAL_PATH_FOR(Source.ANNOTATIONS, Format.YAML)
LOCAL_EXTENSIONS_YAML_FILE = linkml_files.LOCAL_PATH_FOR(Source.EXTENSIONS, Format.YAML)
LOCAL_MODEL_YAML_FILES = (
    LOCAL_METAMODEL_YAML_FILE,
    LOCAL_TYPES_YAML_FILE,
    LOCAL_MAPPINGS_YAML_FILE,
    LOCAL_ANNOTATIONS_YAML_FILE,
    LOCAL_EXTENSIONS_YAML_FILE,
)

# Local location of jsonld and context.jsonld files
LOCAL_METAMODEL_LDCONTEXT_FILE = linkml_files.LOCAL_PATH_FOR(Source.META, Format.JSONLD)
LOCAL_METAMODEL_JSONLD_FILE = linkml_files.LOCAL_PATH_FOR(Source.META, Format.JSON)
LOCAL_TYPES_LDCONTEXT_FILE = linkml_files.LOCAL_PATH_FOR(Source.TYPES, Format.JSONLD)
LOCAL_TYPES_JSONLD_FILE = linkml_files.LOCAL_PATH_FOR(Source.TYPES, Format.JSON)
LOCAL_MAPPINGS_LDCONTEXT_FILE = linkml_files.LOCAL_PATH_FOR(Source.MAPPINGS, Format.JSONLD)
LOCAL_MAPPINGS_JSONLD_FILE = linkml_files.LOCAL_PATH_FOR(Source.MAPPINGS, Format.JSON)

# Local location of metamodel shex file
LOCAL_SHEXJ_FILE_NAME = linkml_files.LOCAL_PATH_FOR(Source.META, Format.SHEXJ)
LOCAL_SHEXC_FILE_NAME = linkml_files.LOCAL_PATH_FOR(Source.META, Format.SHEXC)

# Local location of the metamodel rdf file
LOCAL_RDF_FILE_NAME = linkml_files.LOCAL_PATH_FOR(Source.META, Format.RDF)

# URI for the entire metamodel itself.
METAMODEL_NAME = linkml_files.Source.META.value
METAMODEL_URI = linkml_files.URL_FOR(Source.META, Format.YAML)
METATYPE_NAME = linkml_files.Source.TYPES.value
METATYPE_URI = linkml_files.URL_FOR(Source.TYPES, Format.YAML)
METAMAPPING_NAME = linkml_files.Source.MAPPINGS.value
METAMAPPING_URI = linkml_files.URL_FOR(Source.MAPPINGS, Format.YAML)

# Preferred local name for metamodel elements
METAMODEL_NAMESPACE_NAME = "linkml"
METATYPE_NAMESPACE_NAME = "linkml"
METAMAPPING_NAMESPACE_NAME = "linkml"

# Namespace for metamodel elements
META_BASE_URI = linkml_files.LINKML_NAMESPACE
METAMODEL_NAMESPACE = META_BASE_URI
METATYPE_NAMESPACE = METAMODEL_NAMESPACE
METAMAPPING_NAMESPACE = METAMODEL_NAMESPACE
METAANNOTATIONS_NAMESPACE = METAMODEL_NAMESPACE
METAEXTENSIONS_NAMESPACE = METAMODEL_NAMESPACE
NAMESPACES = (
    METAMODEL_NAMESPACE,
    METATYPE_NAMESPACE,
    METAMAPPING_NAMESPACE,
    METAANNOTATIONS_NAMESPACE,
    METAEXTENSIONS_NAMESPACE,
)

# Metamodel Context URI
METAMODEL_CONTEXT_URI = linkml_files.URL_FOR(Source.META, Format.JSONLD)

# Metamodel ShEx URI
METAMODEL_SHEXJ_URI = linkml_files.URL_FOR(Source.META, Format.SHEXJ)
METAMODEL_SHEXC_URI = linkml_files.URL_FOR(Source.META, Format.SHEXC)

# Metamodel YAML file
METAMODEL_YAML_URI = linkml_files.URL_FOR(Source.META, Format.YAML)

# Biolink model file -- this needs a more official fix
BIOLINK_MODEL_URI = "https://w3id.org/biolink/biolink-model"
BIOLINK_MODEL_PYTHON_LOC = "biolink.model"

TurtleSerializer.roundtrip_prefixes = [""]

REQUESTS_TIMEOUT = 10
"""
Time (in seconds) to wait before failing a network request

See: https://docs.python-requests.org/en/latest/user/advanced/#timeouts
"""



import logging
import sys
from collections.abc import Iterable
from dataclasses import asdict, dataclass, field
from functools import cache
from typing import Any, Union

import click
import jsonschema
from jsonschema.exceptions import best_match
from linkml_runtime.linkml_model import ClassDefinitionName, SchemaDefinition
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.dictutils import as_simple_dict
from linkml_runtime.utils.schemaview import SchemaView
from linkml_runtime.utils.yamlutils import YAMLRoot

from linkml._version import __version__
from linkml.generators.jsonschemagen import JsonSchemaGenerator
from linkml.generators.pythongen import PythonGenerator
from linkml.utils import datautils
from linkml.utils.datavalidator import DataValidator
from linkml.utils.deprecation import deprecation_warning

logger = logging.getLogger(__name__)


class HashableSchemaDefinition(SchemaDefinition):
    def __hash__(self) -> int:
        return hash(self.id)


@cache
def _generate_jsonschema(schema, top_class, closed, include_range_class_descendants):
    deprecation_warning("validators")
    logger.debug("Generating JSON Schema")
    not_closed = not closed
    return JsonSchemaGenerator(
        schema=schema,
        mergeimports=True,
        top_class=top_class,
        not_closed=not_closed,
        include_range_class_descendants=include_range_class_descendants,
    ).generate()


class JsonSchemaDataValidatorError(Exception):
    def __init__(self, validation_messages: list[str]) -> None:
        deprecation_warning("validators")
        super().__init__("\n".join(validation_messages))
        self.validation_messages = validation_messages


@dataclass
class JsonSchemaDataValidator(DataValidator):
    """
    Implementation of DataValidator that wraps jsonschema validation
    """

    include_range_class_descendants: bool = False
    _hashable_schema: Union[str, HashableSchemaDefinition] = field(init=False, repr=False)

    def __setattr__(self, __name: str, __value: Any) -> None:
        deprecation_warning("validators")
        if __name == "schema":
            if isinstance(__value, SchemaDefinition):
                self._hashable_schema = HashableSchemaDefinition(**asdict(__value))
            else:
                self._hashable_schema = __value
        return super().__setattr__(__name, __value)

    def validate_file(self, input: str, format: str = "json", **kwargs):
        # return self.validate_object(obj)
        deprecation_warning("validators")
        pass

    def validate_object(self, data: YAMLRoot, target_class: type[YAMLRoot] = None, closed: bool = True) -> None:
        """
        validates instance data against a schema

        :param data: LinkML instance to be validates
        :param target_class: class in schema to validate against
        :param closed:
        :return:
        """
        deprecation_warning("validators")
        if target_class is None:
            target_class = type(data)
        inst_dict = as_simple_dict(data)
        self.validate_dict(inst_dict, target_class.class_name, closed)

    def validate_dict(self, data: dict, target_class: ClassDefinitionName = None, closed: bool = True) -> None:
        """
        validates instance data against a schema

        :param data: dictionary object
        :param target_class: class in schema to validate against
        :param closed:
        :return:
        """
        deprecation_warning("validators")
        results = list(self.iter_validate_dict(data, target_class, closed))
        if results:
            raise JsonSchemaDataValidatorError(results)

    def iter_validate_dict(
        self, data: dict, target_class_name: ClassDefinitionName = None, closed: bool = True
    ) -> Iterable[str]:
        deprecation_warning("validators")
        if self.schema is None:
            raise ValueError("schema object must be set")
        if target_class_name is None:
            roots = [c.name for c in self.schema.classes.values() if c.tree_root]
            if len(roots) != 1:
                raise ValueError(f"Cannot determine tree root: {roots}")
            target_class_name = roots[0]
        jsonschema_obj = _generate_jsonschema(
            self._hashable_schema, target_class_name, closed, self.include_range_class_descendants
        )
        validator_cls = jsonschema.validators.validator_for(jsonschema_obj, default=jsonschema.Draft7Validator)
        validator = validator_cls(jsonschema_obj, format_checker=validator_cls.FORMAT_CHECKER)
        for error in validator.iter_errors(data):
            best_error = best_match([error])
            # TODO: This should return some kind of standard validation result
            # object, but until that is defined just yield string messages
            yield f"{best_error.message} in {best_error.json_path}"


@click.command(name="jsonschema")
@click.option("--module", "-m", help="Path to python datamodel module")
@click.option(
    "--input-format",
    "-f",
    type=click.Choice(list(datautils.dumpers_loaders.keys())),
    help="Input format. Inferred from input suffix if not specified",
)
@click.option(
    "--target-class",
    "-C",
    help="name of class in datamodel that the root node instantiates",
)
@click.option("--index-slot", "-S", help="top level slot. Required for CSV dumping/loading")
@click.option("--schema", "-s", help="Path to schema specified as LinkML yaml")
@click.option(
    "--exit-on-first-failure/--no-exit-on-first-failure",
    default=False,
    help="Exit after the first validation failure is found. If not specified all validation failures are reported.",
)
@click.option(
    "--include-range-class-descendants/--no-range-class-descendants",
    default=False,
    show_default=False,
    help="""
When handling range constraints, include all descendants of the range class instead of just the range class
""",
)
@click.argument("input")
@click.version_option(__version__, "-V", "--version")
def cli(
    input,
    module,
    target_class,
    input_format=None,
    schema=None,
    index_slot=None,
    exit_on_first_failure=False,
    include_range_class_descendants=False,
) -> None:
    """
    Validates instance data
    """
    deprecation_warning("validators")

    if module is None:
        if schema is None:
            raise Exception("must pass one of module OR schema")
        else:
            python_module = PythonGenerator(schema).compile_module()
    else:
        python_module = compile_python(module)
    if schema is not None:
        sv = SchemaView(schema)
    if target_class is None:
        target_class = datautils.infer_root_class(sv)
    if target_class is None:
        raise Exception("target class not specified and could not be inferred")
    py_target_class = python_module.__dict__[target_class]
    input_format = datautils._get_format(input, input_format)
    loader = datautils.get_loader(input_format)

    inargs = {}
    if datautils._is_xsv(input_format):
        if index_slot is None:
            index_slot = datautils.infer_index_slot(sv, target_class)
            if index_slot is None:
                raise Exception("--index-slot is required for CSV input")
        inargs["index_slot"] = index_slot
        inargs["schema"] = schema
    if datautils._is_rdf_format(input_format):
        inargs["schemaview"] = sv
        inargs["fmt"] = input_format

    try:
        data_as_dict = loader.load_as_dict(source=input, **inargs)
    except NotImplementedError:
        obj = loader.load(source=input, target_class=py_target_class, **inargs)
        data_as_dict = as_simple_dict(obj)

    # Validation
    if schema is None:
        raise Exception("--schema must be passed in order to validate. Suppress with --no-validate")

    validator = JsonSchemaDataValidator(schema, include_range_class_descendants=include_range_class_descendants)
    error_count = 0
    for error in validator.iter_validate_dict(data_as_dict, target_class_name=py_target_class.class_name):
        error_count += 1
        click.echo(click.style("\u2717 ", fg="red") + error)
        if exit_on_first_failure:
            sys.exit(1)

    if not error_count:
        click.echo(click.style("\u2713 ", fg="green") + "No problems found")

    sys.exit(0 if error_count == 0 else 1)


if __name__ == "__main__":
    cli(sys.argv[1:])



import abc
import re
import unicodedata
from dataclasses import dataclass, field
from typing import Optional

from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    EnumDefinition,
    EnumDefinitionName,
    SchemaDefinition,
    SlotDefinition,
    TypeDefinition,
)
from linkml_runtime.utils.formatutils import camelcase, lcamelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml.utils.generator import Generator

SAFE_NAME = str
TYPE_EXPRESSION = str
ANNOTATION = str
PACKAGE = str


@dataclass
class OODocument:
    """
    A collection of one or more OO classes
    """

    name: SAFE_NAME
    package: PACKAGE = None
    source_schema: SchemaDefinition = None
    classes: list["OOClass"] = field(default_factory=lambda: [])
    imports: list[str] = field(default_factory=lambda: [])


@dataclass
class OOField:
    """
    A field belonging to an OO class that corresponds to a LinkML class slot
    """

    name: SAFE_NAME
    range: TYPE_EXPRESSION = None
    default_value: str = None
    annotations: list[ANNOTATION] = field(default_factory=lambda: [])
    source_slot: SlotDefinition = field(default_factory=lambda: [])


@dataclass
class OOClass:
    """
    An object-oriented class
    """

    # ObjectVars
    name: SAFE_NAME
    description: Optional[SAFE_NAME] = None
    is_a: Optional[SAFE_NAME] = None
    mixin: Optional[bool] = None
    abstract: Optional[bool] = None
    mixins: list[SAFE_NAME] = field(default_factory=lambda: [])
    fields: list[OOField] = field(default_factory=lambda: [])
    all_fields: list[OOField] = field(default_factory=lambda: [])
    annotations: list[ANNOTATION] = field(default_factory=lambda: [])
    package: PACKAGE = None
    source_class: ClassDefinition = None


@dataclass
class OOCodeGenerator(Generator):
    # ClassVars
    java_style = True
    visit_all_class_slots = False
    uses_schemaloader = False
    requires_metamodel = False
    schemaview: SchemaView = None

    template_file: str = None
    """Path to template"""

    package: PACKAGE = "example"

    def __post_init__(self):
        # TODO: consider moving up a level
        self.schemaview: SchemaView = SchemaView(self.schema)
        super().__post_init__()

    @abc.abstractmethod
    def serialize(self, directory: str) -> None:
        raise NotImplementedError("Not implemented.")

    @abc.abstractmethod
    def default_value_for_type(self, typ: str) -> str:
        raise NotImplementedError

    @staticmethod
    def get_class_name(cn):
        return camelcase(cn)

    def get_slot_name(self, sn):
        if self.java_style:
            safe_sn = lcamelcase(sn)
        else:
            safe_sn = underscore(sn)
        return safe_sn

    def map_type(self, t: TypeDefinition, required: bool = False) -> str:
        return t.base

    @staticmethod
    def make_multivalued(range: str) -> str:
        return f"List<{range}>"

    @staticmethod
    def replace_invalid_identifier_character(char: str) -> str:
        if char.isalpha() or char.isnumeric() or char == "_":
            return char
        else:
            return underscore(unicodedata.name(char))

    def generate_enum_label(self, value: str) -> str:
        label = underscore(value)
        if label.isidentifier():
            return label
        else:
            # add an underscore if the value starts with a digit
            label = re.sub(r"(?=^\d)", "number_", label)

            safe_label = ""
            for character in label:
                safe_label += self.replace_invalid_identifier_character(character)

            return safe_label

    def generate_enums(self, all_enums: dict[EnumDefinitionName, EnumDefinition]) -> dict:
        # TODO: make an explicit class to represent how an enum is passed to the template
        enums = {}
        for enum_name, enum_original in all_enums.items():
            enum = {"name": camelcase(enum_name), "values": {}}

            if hasattr(enum_original, "description"):
                enum["description"] = enum_original.description

            for pv in enum_original.permissible_values.values():
                if pv.title:
                    label = self.generate_enum_label(pv.title)
                else:
                    label = self.generate_enum_label(pv.text)
                val = {"label": label, "value": pv.text.replace('"', '\\"')}
                if hasattr(pv, "description"):
                    val["description"] = pv.description
                else:
                    val["description"] = None

                enum["values"][label] = val

            enums[enum_name] = enum

        return enums

    def create_documents(self) -> list[OODocument]:
        """
        Currently hardcoded for java-style
        :return:
        """
        sv: SchemaView
        sv = self.schemaview
        docs = []
        for cn in sv.all_classes(imports=False):
            c = sv.get_class(cn)
            safe_cn = camelcase(cn)
            oodoc = OODocument(name=safe_cn, package=self.package, source_schema=sv.schema)
            docs.append(oodoc)
            ooclass = OOClass(
                name=safe_cn,
                description=c.description,
                package=self.package,
                fields=[],
                source_class=c,
            )
            # currently hardcoded for java style, one class per doc
            oodoc.classes = [ooclass]
            if c.mixin:
                ooclass.mixin = c.mixin
            if c.mixins:
                ooclass.mixins = [camelcase(x) for x in c.mixins]
            if c.abstract:
                ooclass.abstract = c.abstract
            if c.is_a:
                ooclass.is_a = self.get_class_name(c.is_a)
                parent_slots = sv.class_slots(c.is_a)
            else:
                parent_slots = []
            for sn in sv.class_slots(cn):
                safe_sn = self.get_slot_name(sn)
                slot = sv.induced_slot(sn, cn)
                range = slot.range
                default_value = "null"

                if range is None:
                    # TODO: schemaview should infer this
                    range = sv.schema.default_range

                if range is None:
                    range = "string"

                if range in sv.all_classes():
                    range = self.get_class_name(range)
                    default_value = "null"
                elif range in sv.all_types():
                    t = sv.get_type(range)
                    range = self.map_type(t, slot.required)
                    if range is None:  # If mapping fails,
                        range = self.map_type(sv.all_types().get("string"))
                elif range in sv.all_enums():
                    range = self.map_type(sv.all_types().get("string"))
                else:
                    raise Exception(f"Unknown range {range}")

                # Set default values for
                if range == "boolean":
                    default_value = "false"
                elif range == "integer":
                    default_value = "0"
                elif range == "String":
                    default_value = '""'

                # TODO:
                #  default_value = default_value_for_type(range)

                if slot.multivalued:
                    range = self.make_multivalued(range)
                    default_value = "List.of()"
                oofield = OOField(
                    name=safe_sn,
                    source_slot=slot,
                    range=range,
                    default_value=default_value,
                )
                if sn not in parent_slots:
                    ooclass.fields.append(oofield)
                ooclass.all_fields.append(oofield)

        return docs



import polars as pl


class CollectionDictLoader:
    def __init__(self, struct_schema, id_col="id", nested_tx=lambda x: x):
        """
        struct_schema:
            the schema of the nested range not including the collection
        id_col: name of column in struct to hold collection key
        nested_tx: transformation to apply to nested dicts
        """
        self.struct_schema = struct_schema
        self.id_col = id_col
        self.nested_tx = nested_tx

    @staticmethod
    def tx_core(collection_dict, id_col="id", nested_tx=lambda x: x):
        """core collection to structs logic"""
        for k, v in collection_dict.items():
            yield {**nested_tx(v), id_col: k}  # Collection key overwrites nested value

    def tx(self, collection_dict):
        """collection_to_structs"""
        return self.tx_core(collection_dict, self.id_col, self.nested_tx)

    def tx_batch(self, series):
        """Process entire series of collection dicts"""
        return pl.Series(
            [list(self.tx(collection_dict)) for collection_dict in series], dtype=pl.List(self.struct_schema)
        )

    def load(self, source_col):
        return pl.col(source_col).map_batches(self.tx_batch, return_dtype=pl.List(self.struct_schema))

    def load_df(self, df, source_col):
        return df.with_columns(self.load(source_col))



import polars as pl

from .model_transform import ModelTransform


class CollectionDictModelTransform(ModelTransform):
    """This class assists in converting a LinkML 'collection dict' inline column
    into a form that is better for representing in a PolaRS dataframe and
    validating with a Pandera model.
    """

    def __init__(self, polars_schema, id_col):
        self.polars_schema = polars_schema
        """A polars schema representing a collection dict column"""

        self.id_col = id_col
        """The ID column in the sense of a LinkML inline collection dict"""

    def transform(self, linkml_collection_dict):
        """Converts a collection dict nested column to a list of dicts.
        { 'A': {...}, 'B': {...}, ... } -> [{'id': 'A', ...}, {'id': 'B', ...}, ...]
        """
        return self._collection_dict_to_list_of_structs(linkml_collection_dict)

    def _collection_dict_to_list_of_structs(self, linkml_collection_dict):
        """Converts a collection dict nested column to a list of dicts.
        { 'A': {...}, 'B': {...}, ... } -> [{'id': 'A', ...}, {'id': 'B', ...}, ...]

        An inefficient conversion (relative to native PolaRS operations)
        from a collection dict form to a dataframe struct column.

        linkml_collection_dict : dict
            A single row entry in a dataframe column (one cell), which itself is a dict.
            The value entries are dicts that get the key added as an id field.
        """
        arr = []
        for k, v in linkml_collection_dict.items():
            if k not in v:
                v[self.id_col] = k
            arr.append(v)
        return arr

    @classmethod
    def prepare_dataframe(cls, data, column_name, nested_cls):
        """Returns just the collection dict column transformed to an inlined list form

        note that this method uses collect and iter_rows so is very inefficient
        """
        id_column = nested_cls.get_id_column_name()
        polars_schema = nested_cls.to_schema()

        collection_dict_transformer = cls(polars_schema, id_column)

        one_column_df = data.lazyframe.select(pl.col(column_name)).collect()

        list_of_structs = [collection_dict_transformer.transform(e) for [e] in one_column_df.iter_rows()]

        return pl.DataFrame(pl.Series(list_of_structs).alias(column_name))

    def explode_unnest_dataframe(self, df, column_name):
        """Filter, explode and unnest for collection dict."""
        return df.lazy().filter(pl.col(column_name).list.len() > 0).explode(column_name).unnest(column_name).collect()



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Callable

import click
from jsonasobj2 import as_dict
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import (
    ClassDefinition,
    ClassDefinitionName,
    ElementName,
    EnumDefinition,
    EnumDefinitionName,
    SchemaDefinition,
    SlotDefinition,
    SlotDefinitionName,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.formatutils import camelcase, lcamelcase, underscore
from linkml_runtime.utils.schema_as_dict import schema_as_yaml_dump
from linkml_runtime.utils.yamlutils import YAMLRoot

ref_map = {
    ClassDefinitionName: ClassDefinition,
    SlotDefinitionName: SlotDefinition,
    EnumDefinitionName: EnumDefinition,
    TypeDefinitionName: TypeDefinition,
}


@dataclass
class SchemaRenamer:
    """
    Renames schema elements
    """

    rename_function_map: dict[type, Callable] = field(default_factory=lambda: {})
    schema: SchemaDefinition = None

    def rename_elements(self, schema: SchemaDefinition) -> SchemaDefinition:
        """
        Rename all elements in the schema using rename_function_map

        :param schema:
        :return:
        """
        schema = SchemaDefinition(**as_dict(schema))
        self.schema = schema
        return self._rename(schema)

    def _rename(self, element: Any, parent_key: ElementName = None) -> Any:
        """
        Renames an individual element plus its children

        Does not modify in place: creates a new element
        :param element:
        :return:
        """
        typ = type(element)
        if element is None:
            return None
        elif isinstance(element, ElementName):
            # element is a *Reference* to a Class, Slot, or other schema element
            if typ == ElementName:
                # cast generic element names, e.g. for ranges
                if element in self.schema.classes:
                    typ = ClassDefinitionName
                elif element in self.schema.enums:
                    typ = EnumDefinitionName
                elif element in self.schema.types:
                    typ = TypeDefinitionName
            typ_deref = ref_map.get(typ, None)
            if typ_deref in self.rename_function_map:
                return typ(self.rename_function_map[typ_deref](element))
            else:
                return element
        elif isinstance(element, dict):
            # element is a dict; keys may be rewired
            new_element = {}
            for k, v in element.items():
                v2 = self._rename(v, k)
                typ = type(v2)
                new_k = k
                if typ in self.rename_function_map:
                    new_k = v2.name
                new_element[new_k] = v2
            return new_element
        elif isinstance(element, list):
            return [self._rename(e) for e in element]
        elif isinstance(element, YAMLRoot):
            # element is a Class, Slot, or other schema element
            new_element = deepcopy(element)
            if typ in self.rename_function_map:
                new_element.name = self.rename_function_map[typ](element.name)
            for k, v in vars(new_element).items():
                setattr(new_element, k, self._rename(v))
            return new_element
        else:
            try:
                element_vars = {k: v for k, v in vars(element).items() if not k.startswith("_")}
                if len(element_vars) == 0:
                    return element
                else:
                    new_element = deepcopy(element)
                    for k, v in vars(new_element).items():
                        setattr(new_element, k, self._rename(v))
                    return new_element
            except TypeError:
                return element


@click.command()
@click.option("-o", "--output", help="path to output schema")
@click.option("-C", "--class-names", help="function for class names")
@click.option("-S", "--slot-names", help="function for slot names")
@click.argument("schema")
def main(schema, output, class_names, slot_names):
    sv = SchemaView(schema)

    def n2f(n: str) -> Callable:
        if n.startswith("camel"):
            return camelcase
        elif n.startswith("lcamel") or n.startswith("lowerpascal"):
            return lcamelcase
        elif n.startswith("snake") or n.startswith("under"):
            return underscore
        elif n.startswith("upper"):
            return lambda x: x.upper()
        elif n.startswith("lower"):
            return lambda x: x.lower()
        else:
            raise NotImplementedError(f"Function: {n}")

    rename_map = {}
    if class_names:
        rename_map[ClassDefinition] = n2f(class_names)
    if slot_names:
        rename_map[SlotDefinition] = n2f(slot_names)
    if not rename_map.keys():
        raise ValueError("No transformations specified")
    renamer = SchemaRenamer(rename_function_map=rename_map)
    rschema = renamer.rename_elements(sv.schema)
    ystr = schema_as_yaml_dump(rschema)
    if output:
        with open(output, "w", encoding="UTF-8") as stream:
            stream.write(ystr)
    else:
        print(ystr)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from .dependency_sorter import DependencySorter

if TYPE_CHECKING:
    from .dataframe_generator import DataframeGenerator


class ClassHandlerBase:
    def __init__(self, generator: "DataframeGenerator"):
        self.generator = generator

    def enum_or_class(self, cn):
        schemaview = self.generator.schemaview

        if cn in schemaview.all_enums():
            return self.generator.schemaview.get_enum(cn, strict=True)

        if cn in schemaview.all_classes():
            return schemaview.get_class(cn, strict=True)

        raise Exception(f"Unknown class or enum {cn}")

    def ordered_classes(self):
        sorter = DependencySorter()

        self.add_dependencies_by_association(sorter)
        self.add_dependencies_for_hierarchy(sorter)
        sorted_class_names = sorter.sort_dependencies()

        ordered = [self.enum_or_class(cn) for cn in sorted_class_names]

        return ordered

    def add_dependencies_for_hierarchy(self, sorter: DependencySorter) -> None:
        """
        Add dependencies based on class hierarchy.
        """
        sv = self.generator.schemaview

        for cn in sv.all_classes():
            sorter.add_dependency(cn, None)
            parents = sv.class_parents(cn)

            for p in parents:
                sorter.add_dependency(cn, p)

    def add_dependencies_by_association(self, sorter: DependencySorter) -> None:
        """
        Add dependencies based on associations.
        """
        sv = self.generator.schemaview

        for cn in sv.all_classes():
            for slot_name in sv.class_slots(cn):
                slot = sv.induced_slot(slot_name, cn)
                if slot.range and slot.range in sv.all_classes() and slot.range not in sv.all_enums():
                    # Only add dependency if slot is inlined or range has no identifier
                    if slot.inlined or slot.inlined_as_list or sv.get_identifier_slot(slot.range) is None:
                        sorter.add_dependency(cn, slot.range)
                    else:
                        sorter.add_dependency(cn, None)



"""
The ``linkml.validator`` package contains a new LinkML validation framework that is more flexible
than the ``linkml.validators`` package. While that package still exists, it may become deprecated
in the future.
"""

import os
from pathlib import Path
from typing import Any, Optional, Union

from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.loaders import yaml_loader

from linkml.validator.loaders import default_loader_for_file
from linkml.validator.plugins import JsonschemaValidationPlugin
from linkml.validator.report import ValidationReport
from linkml.validator.validator import Validator


def _get_default_validator(
    schema: Union[str, dict, Path, SchemaDefinition],
    *,
    strict: bool = False,
) -> Validator:
    try:
        if isinstance(schema, Path):
            schema = str(schema)
        if isinstance(schema, dict):
            schema = SchemaDefinition(**schema)
        elif isinstance(schema, str):
            schema = yaml_loader.load(schema, target_class=SchemaDefinition)

        if not isinstance(schema, SchemaDefinition):
            raise ValueError(f"Schema could not be loaded from {schema}")
    except ValueError as e:
        raise ValueError(f"Invalid schema: {schema}") from e

    validation_plugins = [JsonschemaValidationPlugin(closed=True)]

    return Validator(schema, validation_plugins=validation_plugins, strict=strict)


def validate(
    instance: Any,
    schema: Union[str, dict, SchemaDefinition],
    target_class: Optional[str] = None,
    *,
    strict: bool = False,
) -> ValidationReport:
    """Validate a data instance against a schema

    This function provides a simple interface to do basic validation performed by a JSON Schema
    validator on a single instance. To have more control over the type of validation performed,
    see the :class:`Validator` class.

    :param instance: The instance to validate
    :param schema: The schema used to validate the instance. If a string is
        it will be interpreted as a path, URL, or other loadable location.
        If it is a dict it should be compatible with ``SchemaDefinition``,
        otherwise it should be a ``SchemaDefinition`` instance.
    :param target_class: Name of the class within the schema to validate
        against. If ``None``, the class will be inferred from the schema by
        looking for a class with ``tree_root: true``. Defaults to ``None``.
    :param strict: If ``True``, validation will stop after the first validation
        error is found, Otherwise all validation problems will be reported.
        Defaults to ``False``.
    :raises ValueError: If a valid ``SchemaDefinition`` cannot be constructed
        from the ``schema`` parameter.
    :return: A validation report
    :rtype: ValidationReport
    """
    validator = _get_default_validator(schema, strict=strict)
    return validator.validate(instance, target_class)


def validate_file(
    file: Union[str, bytes, os.PathLike],
    schema: Union[str, dict, SchemaDefinition],
    target_class: Optional[str] = None,
    *,
    strict: bool = False,
) -> ValidationReport:
    """Validate instances loaded from a file against a schema

    This function provides a simple interface to do basic validation performed by a JSON Schema
    validator on instances loaded from a file. Loading is done according to the file's extension.
    Accepted file extensions are: ``.csv``, ``.tsv``, ``.yaml``, ``.yml``, and ``.json``.
    Individual rows of CSV and TSV files are treated as instances to validate. Each document
    within a YAML file is treated as an individual instance to validate. If the top-level of a
    JSON file is an array, each element of the array is treated as an instance to validate.
    Otherwise, if the top-level is an object it is treated as a single instance to validate.

    To have more control over the type of validation performed, see the :class:`Validator` class.

    :param file: Path-like object of the file to be read
    :param schema: The schema used to validate the instance. If a string is
        it will be interpreted as a path, URL, or other loadable location.
        If it is a dict it should be compatible with ``SchemaDefinition``,
        otherwise it should be a ``SchemaDefinition`` instance.
    :param target_class: Name of the class within the schema to validate
        against. If ``None``, the class will be inferred from the schema by
        looking for a class with ``tree_root: true``. Defaults to ``None``.
    :param strict: If ``True``, validation will stop after the first validation
        error is found, Otherwise all validation problems will be reported.
        Defaults to ``False``.
    :return: A validation report
    :rtype: ValidationReport
    """
    loader = default_loader_for_file(file)
    validator = _get_default_validator(schema, strict=strict)
    return validator.validate_source(loader, target_class)


__all__ = ["Validator", "validate", "validate_file"]



"""Linkml linter."""

from __future__ import annotations

import inspect
from collections.abc import Iterable
from copy import deepcopy
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Any

import jsonschema
import yaml
from jsonschema.exceptions import best_match
from jsonschema.protocols import Validator
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SchemaDefinition

from linkml.generators.jsonschemagen import JsonSchemaGenerator

from .. import LOCAL_METAMODEL_YAML_FILE
from .config.datamodel.config import Config, ExtendableConfigs, RuleLevel


@dataclass
class LinterProblem:
    message: str
    level: RuleLevel | None = None
    schema_name: str | None = None
    schema_source: str | None = None
    rule_name: str | None = None


@lru_cache
def get_named_config(name: str) -> dict[str, Any]:
    config_path = str(Path(__file__).parent / f"config/{name}.yaml")
    with open(config_path) as config_file:
        return yaml.safe_load(config_file)


@lru_cache
def get_metamodel_validator() -> Validator:
    meta_json_gen = JsonSchemaGenerator(LOCAL_METAMODEL_YAML_FILE, not_closed=False)
    meta_json_schema = meta_json_gen.generate()
    validator_cls = jsonschema.validators.validator_for(meta_json_schema, default=jsonschema.Draft7Validator)
    return validator_cls(meta_json_schema, format_checker=validator_cls.FORMAT_CHECKER)


def merge_configs(original: dict, other: dict):
    result = deepcopy(original)
    for key, value in other.items():
        if isinstance(value, dict):
            result[key] = merge_configs(result.get(key, {}), value)
        else:
            result[key] = value
    return result


def _format_path_component(value):
    if isinstance(value, int):
        return f"[{value}]"
    return value


def _format_path(path):
    if not path:
        return "<root>"
    return " > ".join(_format_path_component(p) for p in path)


class Linter:
    def __init__(self, config: dict[str, Any] = {}) -> None:
        default_config = deepcopy(get_named_config("default"))
        merged_config = config
        if config.get("extends") == ExtendableConfigs.recommended.text:
            recommended_config = deepcopy(get_named_config(ExtendableConfigs.recommended.text))
            merged_config = merge_configs(recommended_config, merged_config)
        merged_config = merge_configs(default_config, merged_config)
        self.config = Config(**merged_config)

        from . import rules

        self._rules_map = dict(
            [
                (cls.id, cls)
                for _, cls in inspect.getmembers(rules, inspect.isclass)
                if issubclass(cls, rules.LinterRule)
            ]
        )

    @staticmethod
    def validate_schema(schema_path: str):
        with open(schema_path) as schema_file:
            schema = yaml.safe_load(schema_file)

        validator = get_metamodel_validator()
        for err in validator.iter_errors(schema):
            best_err = best_match([err])
            message = f"In {_format_path(best_err.absolute_path)}: {best_err.message}"
            if best_err.context:
                message += f" ({', '.join(e.message for e in best_err.context)})"
            yield LinterProblem(
                rule_name="valid-schema",
                message=message,
                level=RuleLevel(RuleLevel.error),
                schema_source=schema,
            )

    def lint(
        self,
        schema: str | SchemaDefinition,
        fix: bool = False,
        validate_schema: bool = False,
        validate_only: bool = False,
    ) -> Iterable[LinterProblem]:
        if (validate_schema or validate_only) and isinstance(schema, str):
            yield from self.validate_schema(schema)

        if validate_only:
            return

        try:
            schema_view = SchemaView(schema)
        except Exception:
            if not validate_schema:
                yield LinterProblem(
                    message="File is not a valid LinkML schema. Use --validate for more details.",
                    level=RuleLevel(RuleLevel.error),
                    schema_source=(schema if isinstance(schema, str) else None),
                )
            return

        for rule_id, rule_config in self.config.rules.__dict__.items():
            rule_cls = self._rules_map.get(rule_id, None)
            if rule_cls is None:
                raise ValueError("Unknown rule id: " + rule_id)

            if str(rule_config.level) is RuleLevel.disabled.text:
                continue

            rule = rule_cls(rule_config)

            for problem in rule.check(schema_view, fix=fix):
                problem.level = rule.config.level
                problem.rule_name = rule.id
                problem.schema_name = schema_view.schema.name
                if isinstance(schema, str):
                    problem.schema_source = schema
                yield problem

        if fix and schema_view.schema.source_file:
            yaml_dumper.dump(schema_view.schema, schema_view.schema.source_file)



from collections import defaultdict
from typing import IO, Any, Optional

import click

from linkml.linter.config.datamodel.config import RuleLevel

from ..linter import LinterProblem
from .formatter import Formatter


def plural(word: str, count: int):
    return word + ("s" if count != 1 else "")


class TerminalFormatter(Formatter):
    def __init__(self, file: Optional[IO[Any]] = None, verbose: bool = False) -> None:
        super().__init__(file)
        self.verbose = verbose
        self.problem_counts = defaultdict(int)
        self.current_schema = None

    def start_schema(self, name: str):
        self.current_schema = name
        if self.verbose:
            self.write(click.style(name, underline=True))

    def handle_problem(self, problem: LinterProblem):
        key = self.current_schema
        if not self.verbose and key not in self.problem_counts:
            self.write(click.style(key, underline=True))

        self.problem_counts[key] += 1

        formatted = click.style(
            "  " + str(problem.level).ljust(9),
            fg="yellow" if str(problem.level) is RuleLevel.warning.text else "red",
        )
        formatted += problem.message
        if problem.rule_name:
            formatted += "  " + click.style(f"({problem.rule_name})", dim=True)
        self.write(formatted)

    def end_schema(self):
        if self.verbose or self.current_schema in self.problem_counts:
            self.write("")

    def end_report(self):
        total_problems = sum(self.problem_counts.values())
        if total_problems > 0:
            problem_schemas = len(self.problem_counts)
            self.write(
                click.style("\u2716", fg="red")
                + " Found "
                + str(total_problems)
                + " "
                + plural("problem", total_problems)
                + " in "
                + str(problem_schemas)
                + " "
                + plural("schema", problem_schemas)
            )
        else:
            self.write(click.style("\u2713", fg="green") + " No problems found")



"""Generate ShEx definition of a model"""

import os
import urllib.parse as urlparse
from dataclasses import dataclass, field
from typing import Optional, Union

import click
from jsonasobj import as_json as as_json_1
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ElementName,
    EnumDefinition,
    SlotDefinition,
    SlotDefinitionName,
    TypeDefinition,
)
from linkml_runtime.linkml_model.types import SHEX
from linkml_runtime.utils.formatutils import camelcase, sfx
from linkml_runtime.utils.metamodelcore import URIorCURIE
from rdflib import OWL, RDF, XSD, Graph, Namespace
from ShExJSG import ShExC
from ShExJSG.SchemaWithContext import Schema
from ShExJSG.ShExJ import IRIREF, EachOf, NodeConstraint, Shape, ShapeOr, TripleConstraint

from linkml import METAMODEL_NAMESPACE, METAMODEL_NAMESPACE_NAME
from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class ShExGenerator(Generator):
    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.2"
    valid_formats = ["shex", "json", "rdf"]
    file_extension = "shex.rdf"
    visit_all_class_slots = False
    uses_schemaloader = True

    # ObjectVars
    shex: Schema = field(default_factory=lambda: Schema())  # ShEx Schema being generated
    shapes: list = field(default_factory=lambda: [])
    shape: Optional[Shape] = None  # Current shape being defined
    list_shapes: list[IRIREF] = field(default_factory=lambda: [])  # Shapes that have been defined as lists

    def __post_init__(self):
        super().__post_init__()

        if METAMODEL_NAMESPACE_NAME not in self.namespaces:
            self.namespaces[METAMODEL_NAMESPACE_NAME] = METAMODEL_NAMESPACE
        self.meta = Namespace(
            self.namespaces.join(self.namespaces[METAMODEL_NAMESPACE_NAME], "")
        )  # URI for the metamodel
        self.base = Namespace(self.namespaces.join(self.namespaces._base, ""))  # Base URI for what is being modeled

    def generate_header(self) -> str:
        out = f"# metamodel_version: {self.schema.metamodel_version}\n"
        if self.schema.version:
            out += f"# version: {self.schema.version}\n"
        return out

    def visit_schema(self, **_):
        # Adjust the schema context to include the base model URI
        context = self.shex["@context"]
        self.shex["@context"] = [context, {"@base": self.namespaces._base}]
        # Emit all of the type definitions
        for typ in self.schema.types.values():
            model_uri = self._class_or_type_uri(typ)
            if typ.uri:
                typ_type_uri = self.namespaces.uri_for(typ.uri)
                if typ_type_uri in (XSD.anyURI, SHEX.iri):
                    self.shapes.append(NodeConstraint(id=model_uri, nodeKind="iri"))
                elif typ_type_uri == SHEX.nonLiteral:
                    self.shapes.append(NodeConstraint(id=model_uri, nodeKind="nonliteral"))
                else:
                    self.shapes.append(NodeConstraint(id=model_uri, datatype=self.namespaces.uri_for(typ.uri)))
            else:
                typeof_uri = self._class_or_type_uri(typ.typeof)
                self.shapes.append(Shape(id=model_uri, expression=typeof_uri))
        if self.format != "json":
            return self.generate_header()

    def visit_class(self, cls: ClassDefinition) -> bool:
        self.shape = Shape()

        # Start with all the parent classes, mixins and appytos
        struct_ref_list = [cls.is_a] if cls.is_a else []
        struct_ref_list += cls.mixins
        if cls.name in self.synopsis.applytorefs:
            for applier in self.synopsis.applytorefs[cls.name].classrefs:
                struct_ref_list.append(applier)
        for sr in struct_ref_list:
            self._add_constraint(self._class_or_type_uri(sr, "_tes"))
            self._add_constraint(self._type_arc(self.schema.classes[sr].class_uri, opt=True))
        return True

    def end_class(self, cls: ClassDefinition) -> None:
        # On entry self.shape contains all of the triple expressions that define the body of the shape

        # Finish off the shape definition itself

        # If there is nothing yet, we're at the very root of things.  Add in a final catch-all for any additional
        # type arcs.  NOTE: Here is where you can sink other things as well if you want to ignore categories of things
        if self.shape.expression is None:
            self._add_constraint(TripleConstraint(predicate=RDF.type, min=0, max=-1))
        self.shape.expression.id = self._class_or_type_uri(cls, "_tes")
        self.shape.expression = EachOf(
            expressions=[
                self.shape.expression,
                self._type_arc(cls.class_uri, not bool(self.class_identifier(cls))),
            ]
        )
        self.shape.closed = not (cls.abstract or cls.mixin)

        # If this class has subtypes, define the class as the union of its subtypes and itself (if not abstract)
        if cls.name in self.synopsis.isarefs:
            childrenExprs = []
            for child_classname in sorted(list(self.synopsis.isarefs[cls.name].classrefs)):
                childrenExprs.append(self._class_or_type_uri(child_classname))
            if not (cls.mixin or cls.abstract) or len(childrenExprs) == 1:
                childrenExprs.insert(0, self.shape)
                self.shapes.append(ShapeOr(id=self._class_or_type_uri(cls), shapeExprs=childrenExprs))
            else:
                self.shapes.append(ShapeOr(id=self._class_or_type_uri(cls), shapeExprs=childrenExprs))
                self.shape.id = self._class_or_type_uri(cls, "_struct")
                self.shapes.append(self.shape)
        else:
            self.shape.id = self._class_or_type_uri(cls)
            self.shapes.append(self.shape)

    def visit_class_slot(
        self,
        cls: ClassDefinition,
        aliased_slot_name: SlotDefinitionName,
        slot: SlotDefinition,
    ) -> None:
        if not (slot.identifier or slot.abstract or slot.mixin):
            constraint = TripleConstraint()
            self._add_constraint(constraint)
            constraint.predicate = self.namespaces.uri_for(slot.slot_uri)
            constraint.min = int(bool(slot.required))
            constraint.max = 1 if not slot.multivalued else -1
            if slot.range in self.schema.enums:
                # Handle permissible values from enums
                enum = self.schema.enums[slot.range]
                values = []
                for value in enum.permissible_values.values():
                    if value.meaning:
                        values.append(self.namespaces.uri_for(value.meaning))
                    else:
                        value_uri = f"{self._class_or_type_uri(enum.name)}#{urlparse.quote(value.text)}"
                        values.append(value_uri)
                if values:
                    node_constraint = NodeConstraint(
                        # id=self._class_or_type_uri(slot.range),
                        values=values,
                    )
                    constraint.valueExpr = node_constraint
            else:
                constraint.valueExpr = self._class_or_type_uri(slot.range)

    def end_schema(self, output: Optional[str] = None, **_) -> str:
        self.shex.shapes = self.shapes if self.shapes else [Shape()]
        shex = as_json_1(self.shex)
        if self.format == "rdf":
            g = Graph()
            g.parse(data=shex, format="json-ld", version="1.1")
            g.bind("owl", OWL)
            shex = g.serialize(format="turtle")
        elif self.format == "shex":
            g = Graph()
            self.namespaces.load_graph(g)
            shex = str(ShExC(self.shex, base=sfx(self.namespaces._base), namespaces=g))

        if output:
            with open(output, "w", encoding="UTF-8") as outf:
                outf.write(shex)
        return shex

    def _class_or_type_uri(
        self,
        item: Union[TypeDefinition, ClassDefinition, ElementName],
        suffix: Optional[str] = "",
    ) -> URIorCURIE:
        # TODO: enums - figure this out
        if isinstance(item, (TypeDefinition, ClassDefinition, EnumDefinition)):
            cls_or_type = item
        else:
            cls_or_type = self.class_or_type_for(item)
        return self.namespaces.uri_for(
            self.namespaces.uri_or_curie_for(
                self.schema_defaults[cls_or_type.from_schema],
                camelcase(cls_or_type.name) + suffix,
            )
        )

    def _slot_uri(self, name: str, suffix: Optional[str] = "") -> URIorCURIE:
        slot = self.schema.slots[name]
        return self.namespaces.uri_for(
            self.namespaces.uri_or_curie_for(self.schema_defaults[slot.from_schema], camelcase(name) + suffix)
        )

    def _add_constraint(self, constraint) -> None:
        # No constraints
        if not self.shape.expression:
            self.shape.expression = constraint
        # One constraint
        elif not isinstance(self.shape.expression, EachOf):
            self.shape.expression = EachOf(expressions=[self.shape.expression, constraint])
        # Two or more constraints
        else:
            self.shape.expression.expressions.append(constraint)

    def _type_arc(self, target: URIorCURIE, opt: bool = False) -> TripleConstraint:
        return TripleConstraint(
            predicate=RDF.type,
            valueExpr=NodeConstraint(values=[IRIREF(self.namespaces.uri_for(target))]),
            min=0 if opt else 1,
        )


@shared_arguments(ShExGenerator)
@click.command(name="shex")
@click.option("-o", "--output", help="Output file name")
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **args):
    """Generate a ShEx Schema for a  LinkML model"""
    print(ShExGenerator(yamlfile, **args).serialize(**args))


if __name__ == "__main__":
    cli()



"""
Utilities for deprecating functionality and dependencies.

- Emitting DeprecationWarnings
- Tracking deprecated and removed in versions
- Fail tests when something marked as removed_in is still present in the specified version

Initial draft for deprecating Pydantic 1, to make more general, needs
- function wrapper version
- ...

To deprecate something:

- Create a :class:`.Deprecation` object within the `DEPRECATIONS` tuple
- Use the :func:`.deprecation_warning` function wherever the deprecated feature would be used to emit the warning

"""

import re
import warnings
from dataclasses import dataclass
from importlib.metadata import version
from typing import Optional

# Stolen from https://github.com/pypa/packaging/blob/main/src/packaging/version.py
# Updated to include major, minor, and patch versions
PEP440_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""
PEP440 = re.compile(r"^\s*" + PEP440_PATTERN + r"\s*$", re.VERBOSE | re.IGNORECASE)


@dataclass
class SemVer:
    """
    Representation of semantic version that supports inequality comparisons.

    .. note::

        The inequality methods _only_ test the numeric major, minor, and patch components
        of the version - ie. they do not evaluate the prerelease versions as described in the semver
        spec. This is not intended to be a general SemVer inequality calculator, but used
        only for testing deprecations

    """

    major: int = 0
    minor: int = 0
    patch: int = 0
    epoch: Optional[int] = None
    pre: Optional[str] = None
    pre_l: Optional[str] = None
    pre_n: Optional[str] = None
    post: Optional[str] = None
    post_n1: Optional[str] = None
    post_l: Optional[str] = None
    post_n2: Optional[str] = None
    dev: Optional[str] = None
    dev_l: Optional[str] = None
    dev_n: Optional[str] = None
    local: Optional[str] = None

    def __post_init__(self):
        self.major = int(self.major)
        self.minor = int(self.minor)
        self.patch = int(self.patch)

    @classmethod
    def from_str(cls, v: str) -> Optional["SemVer"]:
        """
        Create a SemVer from a string using `PEP 440 <https://peps.python.org/pep-0440/>`_
        syntax.

        Examples:

            .. code-block:: python

                >>> version = SemVer.from_str("v0.1.0")
                >>> print(version)
                0.1.0

        """
        match = PEP440.search(v)
        if match is None:
            return None
        return SemVer(**match.groupdict())

    @classmethod
    def from_package(cls, package: str) -> "SemVer":
        """Get semver from package name"""
        v = version(package)
        return SemVer.from_str(v)

    def __eq__(self, other: "SemVer"):
        return self.major == other.major and self.minor == other.minor and self.patch == other.patch

    def __lt__(self, other: "SemVer"):
        # fall through each if elif only if version component is equal
        for field in ("major", "minor", "patch"):
            if getattr(self, field) < getattr(other, field):
                return True
            elif getattr(self, field) > getattr(other, field):
                return False

        # otherwise, equal (which is False)
        return False

    def __gt__(self, other: "SemVer"):
        return not (self < other) and not (self == other)

    def __le__(self, other: "SemVer"):
        return (self < other) or (self == other)

    def __ge__(self, other: "SemVer"):
        return (self > other) or (self == other)

    def __str__(self) -> str:
        return ".".join([str(item) for item in [self.major, self.minor, self.patch]])


@dataclass
class Deprecation:
    """
    Parameterization of a deprecation.
    """

    name: str
    """Shorthand, unique name used to refer to this deprecation"""
    message: str
    """Message to be displayed explaining the deprecation"""
    deprecated_in: SemVer
    """Version that the feature was deprecated in"""
    removed_in: Optional[SemVer] = None
    """Version that the feature will be removed in"""
    recommendation: Optional[str] = None
    """Recommendation about what to do to replace the deprecated behavior"""
    issue: Optional[int] = None
    """GitHub version describing deprecation"""

    def __post_init__(self):
        if self.deprecated_in is not None and isinstance(self.deprecated_in, str):
            self.deprecated_in = SemVer.from_str(self.deprecated_in)
        if self.removed_in is not None and isinstance(self.removed_in, str):
            self.removed_in = SemVer.from_str(self.removed_in)

    def __str__(self) -> str:
        msg = f"[{self.name}] "
        if self.removed:
            msg += "REMOVED"
        elif self.deprecated:
            msg += "DEPRECATED"

        msg += f"\n{self.message}"
        msg += f"\nDeprecated In: {str(self.deprecated_in)}"
        if self.removed_in is not None:
            msg += f"\nRemoved In: {str(self.removed_in)}"
        if self.recommendation is not None:
            msg += f"\nRecommendation: {self.recommendation}"
        if self.issue is not None:
            msg += f"\nSee: https://github.com/linkml/linkml/issues/{self.issue}"
        return msg

    @property
    def deprecated(self) -> bool:
        return SemVer.from_package("linkml") >= self.deprecated_in

    @property
    def removed(self) -> bool:
        if self.removed_in is None:
            return False
        return SemVer.from_package("linkml") >= self.removed_in

    def warn(self, **kwargs):
        if self.deprecated:
            warnings.warn(message=str(self), category=DeprecationWarning, stacklevel=3, **kwargs)


DEPRECATIONS = (
    Deprecation(
        name="pydanticgen-v1",
        deprecated_in=SemVer.from_str("1.7.5"),
        removed_in=SemVer.from_str("1.10.0"),
        message="Support for generating Pydantic v1.*.* models with pydanticgen is deprecated",
        recommendation="Migrate any existing models to Pydantic v2",
        issue=1925,
    ),
    Deprecation(
        name="pydantic-v1",
        deprecated_in=SemVer.from_str("1.7.5"),
        removed_in=SemVer.from_str("1.10.0"),
        message=(
            "LinkML will set a dependency of pydantic>=2 and become incompatible "
            "with packages with pydantic<2 as a runtime dependency"
        ),
        recommendation="Update dependent packages to use pydantic>=2",
        issue=1925,
    ),
    Deprecation(
        name="validators",
        deprecated_in=SemVer.from_str("1.8.6"),
        removed_in=SemVer.from_str("1.10.0"),
        message=(
            "linkml.validators and linkml.utils.validation are the older versions "
            "of linkml.validator and have unmaintained, duplicated functionality"
        ),
        recommendation="Update to use linkml.validator",
    ),
    Deprecation(
        name="gen-markdown",
        # the last update to any code in markdowngen.py was in v1.9.1
        # we can start considering it as deprecated from this version
        deprecated_in=SemVer.from_str("1.9.1"),
        removed_in=SemVer.from_str("1.10.0"),
        message=(
            "gen-markdown has been deprecated in favor of gen-doc, which is the new "
            "de-facto generator for generating markdown documentation files (with "
            "embedded mermaid class diagrams) from a LinkML schema"
        ),
        recommendation="Update to use `gen-doc`",
    ),
    Deprecation(
        name="gen-yuml",
        # the last update to any code in yumlgen.py was in v1.8.7
        # we can start considering it as deprecated from this version
        deprecated_in=SemVer.from_str("1.8.7"),
        removed_in=SemVer.from_str("1.10.0"),
        message=(
            "gen-yuml has been deprecated and is no longer supported. "
            "It has been replaced by more feature-rich diagram generators such as "
            "gen-doc (embedded mermaid diagrams), gen-plantuml, gen-mermaid-class-diagram, and gen-erdiagram."
        ),
        recommendation="Update to use `gen-doc`, `gen-plantuml`, `gen-mermaid-class-diagram`, or `gen-erdiagram`",
    ),
)  # type: tuple[Deprecation, ...]

EMITTED = set()  # type: set[str]


def deprecation_warning(name: str):
    """
    Call this with the name of the deprecation object wherever the deprecated functionality will be used

    This function will

    - emit a warning if the current version is greater than ``deprecated_in``
    - log that the deprecated feature was accessed in ``EMITTED`` for testing deprecations and muting warnings

    """
    global DEPRECATIONS
    global EMITTED

    dep = [dep for dep in DEPRECATIONS if dep.name == name]
    if len(dep) == 1:
        dep = dep[0]
    elif len(dep) > 1:
        raise RuntimeError(f"Duplicate deprecations found with name {name}")
    else:
        EMITTED.add(name)
        return

    if dep.name not in EMITTED:
        dep.warn()

    EMITTED.add(name)



import logging
from typing import TYPE_CHECKING, Optional

from linkml_runtime.linkml_model.meta import ClassDefinitionName, SlotDefinition

if TYPE_CHECKING:
    from .dataframe_generator import DataframeGenerator

from linkml.utils.helpers import get_range_associated_slots

from .render_adapters.dataframe_field import DataframeField

logger = logging.getLogger(__file__)


class SlotHandlerBase:
    """
    Prior to rendering the dataframe schema, this class provides
    and adapter between the LinkML model and schema view
    and the rendering engine.
    """

    def __init__(self, generator: "DataframeGenerator"):
        self.generator = generator

    LINKML_ANY_CURIE = "linkml:Any"

    # constants used to render the schema
    # these will be moved to a dialect-specific place
    ANY_RANGE_STRING = "Object"
    CLASS_RANGE_STRING = "Struct"
    SIMPLE_DICT_RANGE_STRING = "Struct"
    ENUM_RANGE_STRING = "Enum"

    # association form flags used for rendering decisions
    FORM_INLINED_DICT = "inlined_dict"
    FORM_INLINED_LIST_DICT = "inlined_list_dict"
    FORM_INLINED_COLLECTION_DICT = "inline_collection_dict"
    FORM_INLINED_SIMPLE_DICT = "simple_dict"
    FORM_MULTIVALUED_FOREIGN_KEY = "list_foreign_key"
    FORM_FOREIGN_KEY = "foreign_key"
    FORM_ERROR = "error"

    # When nested inlining is done, the Pandera validator needs a specific range
    INLINED_FORM_RANGE_PANDERA = {
        FORM_INLINED_SIMPLE_DICT: SIMPLE_DICT_RANGE_STRING,
        FORM_INLINED_LIST_DICT: CLASS_RANGE_STRING,
        FORM_INLINED_COLLECTION_DICT: CLASS_RANGE_STRING,
        FORM_INLINED_DICT: CLASS_RANGE_STRING,
        FORM_ERROR: None,
    }

    def is_multivalued(self, slot):
        return "multivalued" in slot and slot.multivalued is True

    _INTERNAL_INLINED_FORM = {
        # INLINED, INLINED_AS_LIST, MULTIVALUED,
        (False, False, False): FORM_FOREIGN_KEY,
        (False, False, True): FORM_MULTIVALUED_FOREIGN_KEY,
        (False, True, False): FORM_INLINED_LIST_DICT,
        (False, True, True): FORM_INLINED_LIST_DICT,
        (True, False, False): FORM_INLINED_DICT,
        (True, False, True): FORM_INLINED_COLLECTION_DICT,
        (True, None, True): FORM_INLINED_DICT,
        (True, True, False): FORM_INLINED_DICT,
        (True, True, True): FORM_INLINED_LIST_DICT,
    }

    def get_identifier_or_key_slot(self, cn: ClassDefinitionName) -> Optional[SlotDefinition]:
        sv = self.generator.schemaview
        id_slot = sv.get_identifier_slot(cn)
        if id_slot:
            return id_slot
        else:
            for s in sv.class_induced_slots(cn):
                if s.key:
                    return s
            return None

    def calculate_inlined_form(self, slot: SlotDefinition) -> str:
        is_multivalued = self.is_multivalued(slot)
        internal_inlined_form_key = ((slot.inlined is True), (slot.inlined_as_list is True), is_multivalued)
        logger.info(f"Inlined form key: {internal_inlined_form_key}")
        internal_inlined_form = self._INTERNAL_INLINED_FORM.get(internal_inlined_form_key, SlotHandlerBase.FORM_ERROR)

        if internal_inlined_form == SlotHandlerBase.FORM_INLINED_COLLECTION_DICT:
            if self.get_identifier_or_key_slot(slot.range) is None:
                internal_inlined_form = SlotHandlerBase.FORM_INLINED_LIST_DICT

        if self.calculate_simple_dict(slot) is not None:
            return SlotHandlerBase.FORM_INLINED_SIMPLE_DICT

        return internal_inlined_form

    def calculate_simple_dict(self, slot: SlotDefinition):
        """slot is the container for the simple dict slot"""

        (_, range_simple_dict_value_slot, _) = get_range_associated_slots(self.generator.schemaview, slot.range)

        return range_simple_dict_value_slot

    def handle_none_slot(self, slot) -> str:
        range = self.generator.schema.default_range  # need to figure this out, set at the beginning?
        if range is None:
            range = "str"

        return range

    def handle_class_slot(self, slot, range: str) -> str:
        range_info = self.generator.schemaview.all_classes().get(range)

        if range_info["class_uri"] == SlotHandlerBase.LINKML_ANY_CURIE:
            range = SlotHandlerBase.ANY_RANGE_STRING
        else:
            inlined_form = self.calculate_inlined_form(slot)

            if inlined_form == SlotHandlerBase.FORM_INLINED_COLLECTION_DICT:
                logger.warning(
                    f"Slot {slot.name} uses inlined dictionary form,"
                    "which may be less efficient than inlined as list form with the current implementation."
                )
            elif inlined_form == SlotHandlerBase.FORM_INLINED_SIMPLE_DICT:
                logger.warning(
                    f"Slot {slot.name} uses inlined simple dictionary form. Support is incomplete "
                    "and performance is less efficient than inlined as list form with the current implementation."
                )

            if inlined_form in (SlotHandlerBase.FORM_MULTIVALUED_FOREIGN_KEY, SlotHandlerBase.FORM_FOREIGN_KEY):
                logger.warning(f"Foreign key not implemented for slot {slot.name}")
                range = f"ID_TYPES['{self.generator.get_class_name(range)}']"
            else:
                # TODO: make these setters
                slot.annotations["reference_class"] = self.generator.get_class_name(range)
                slot.annotations["inline_form"] = inlined_form

                range = SlotHandlerBase.INLINED_FORM_RANGE_PANDERA[inlined_form]

                if inlined_form == SlotHandlerBase.FORM_INLINED_SIMPLE_DICT:
                    self.set_simple_dict_inline_details_annotation(slot)
                elif inlined_form in [SlotHandlerBase.FORM_INLINED_LIST_DICT]:
                    range = self.generator.make_multivalued(range)

        return range

    def set_simple_dict_inline_details_annotation(self, slot):
        """Extra metadata is to help with the simple dict case"""
        (range_id_slot, range_simple_dict_value_slot, _) = get_range_associated_slots(  # range_required_slots,
            self.generator.schemaview, slot.range
        )

        simple_dict_id = range_id_slot.name
        other_slot = range_simple_dict_value_slot.name
        slot.annotations["inline_details"] = {"id": simple_dict_id, "other": other_slot}

    def handle_non_inlined_class_slot(self, slot, range: str) -> str:
        """non-inlined class slots have been temporarily removed but this will be needed to support them"""
        return f"ID_TYPES['{self.generator.get_class_name(range)}']"

    def handle_type_slot(self, slot, range: str) -> str:
        del slot  # unused for now

        t = self.generator.schemaview.all_types().get(range)
        range = self.generator.map_type(t)

        return range

    def handle_enum_slot(self, slot, range: str) -> str:
        enum_definition = self.generator.schemaview.all_enums().get(range)
        range = SlotHandlerBase.ENUM_RANGE_STRING
        slot.annotations["permissible_values"] = self.generator.enum_handler.get_enum_permissible_values(
            enum_definition
        )

        return range

    def handle_multivalued_slot(self, slot, range: str) -> str:
        if (slot.inlined_as_list is True and self.is_multivalued(slot)) or (
            slot.inlined is True and slot.inlined_as_list is True and self.is_multivalued(slot)
        ):
            range = self.generator.make_multivalued(range)

        return range

    def handle_slot(self, cn: str, sn: str):
        safe_sn = self.generator.get_slot_name(sn)
        slot = self.generator.schemaview.induced_slot(sn, cn)
        range = slot.range

        if slot.alias is not None:
            safe_sn = self.generator.get_slot_name(slot.alias)

        if range is None:
            range = self.handle_none_slot(slot)
        elif range in self.generator.schemaview.all_classes():
            range = self.handle_class_slot(slot, range)
        elif range in self.generator.schemaview.all_types():
            range = self.handle_type_slot(slot, range)
            if self.is_multivalued(slot):
                range = self.generator.make_multivalued(range)
        elif range in self.generator.schemaview.all_enums():
            range = self.handle_enum_slot(slot, range)
            range = self.handle_multivalued_slot(slot, range)
        else:
            raise Exception(f"Unknown range {range}")

        return DataframeField(
            name=safe_sn,
            source_slot=slot,
            range=range,
        )



import logging
from dataclasses import dataclass
from pathlib import Path

import click
from linkml_runtime.utils.schemaview import SchemaView
from openpyxl import Workbook
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.datavalidation import DataValidation

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class ExcelGenerator(Generator):
    # ClassVars
    generatorname = Path(__file__).name
    generatorversion = "0.1.1"
    valid_formats = ["xlsx"]
    uses_schemaloader = False
    requires_metamodel = False

    split_workbook_by_class: bool = False
    include_mixins: bool = False

    def __post_init__(self) -> None:
        super().__post_init__()
        self.logger = logging.getLogger(__name__)
        self.schemaview = SchemaView(self.schema)

    @staticmethod
    def create_workbook(workbook_path: Path) -> Workbook:
        """
        Creates an Excel workbook using the openpyxl library and returns it.

        :param workbook_path: Path of the workbook to be created.
        :return: An openpyxl Workbook object representing the newly created workbook.
        """
        workbook = Workbook()
        workbook.save(workbook_path)
        return workbook

    def create_workbook_and_worksheets(self, output_path: Path, classes: list[str]) -> None:
        """
        Creates a workbook with worksheets for each class.

        :param output_path: The path where the workbook should be created.
        :param classes: List of class names for which worksheets should be created.
        """
        workbook = self.create_workbook(output_path)
        workbook.remove(workbook.active)
        sv = self.schemaview

        for cls_name in classes:
            workbook.create_sheet(cls_name)

            # Add columns to the worksheet for the current class
            slots = [s.name for s in sv.class_induced_slots(cls_name, self.mergeimports)]
            self.add_columns_to_worksheet(workbook, cls_name, slots)
            workbook.save(output_path)

            # Add enum validation for columns with enum types
            enum_list = list(sv.all_enums(imports=self.mergeimports).keys())
            for s in sv.class_induced_slots(cls_name, self.mergeimports):
                if s.range in enum_list:
                    pv_list = list(sv.get_enum(s.range).permissible_values.keys())

                    # Data Validation formula to be applied to the column and
                    # which will be used to create the dropdown
                    dv_formula = f'"{",".join(pv_list)}"'

                    # Check if the total length of the data validation formula
                    # including the separators is <= 255 characters
                    enum_length = len(dv_formula)
                    if enum_length <= 255:
                        self.column_enum_validation(workbook, cls_name, s.name, dv_formula)
                    else:
                        self.logger.warning(
                            f"'{s.range}' has permissible values with total "
                            "length > 255 characters. Dropdowns may not work properly "
                            f"in {output_path}"
                        )
                workbook.save(output_path)

        workbook.save(output_path)
        if self.split_workbook_by_class:
            self.logger.info(f"The Excel workbooks have been written to {output_path}")

    @staticmethod
    def add_columns_to_worksheet(workbook: Workbook, worksheet_name: str, sheet_headings: list[str]) -> None:
        """
        Get a worksheet by name and add a column to it in an existing workbook.

        :param workbook: The workbook to which the worksheet should be added.
        :param worksheet_name: Name of the worksheet to add the column to.
        :param column_data: List of data to populate the column with.
        """
        # Get the worksheet by name
        worksheet = workbook[worksheet_name]

        # Add the headings to the worksheet
        for i, heading in enumerate(sheet_headings):
            worksheet.cell(row=1, column=i + 1, value=heading)

    @staticmethod
    def column_enum_validation(
        workbook: Workbook,
        worksheet_name: str,
        column_name: str,
        dv_formula: str,
    ) -> None:
        """
        Get worksheet by name and add a dropdown to a specific column in it
        based on a list of values.

        :param workbook: The workbook to which the worksheet should be added.
        :param worksheet_name: Name of the worksheet to add the column dropdown to.
        :param column_name: Name of the worksheet column to add the dropdown to.
        :param dv_formula: Validation formula (as a literal) to be used for the dropdown.
        """
        worksheet = workbook[worksheet_name]

        column_list = [cell.value for cell in worksheet[1]]
        column_number = column_list.index(column_name) + 1
        column_letter = get_column_letter(column_number)

        # Create the data validation object and set the dropdown values
        dv = DataValidation(type="list", formula1=dv_formula, allow_blank=True)

        worksheet.add_data_validation(dv)

        dv.add(f"{column_letter}2:{column_letter}1048576")

    def serialize(self, **kwargs) -> str:
        sv = self.schemaview

        if self.include_mixins:
            classes_to_process = [
                cls_name for cls_name, cls in sv.all_classes(imports=self.mergeimports).items() if not cls.abstract
            ]
        else:
            classes_to_process = [
                cls_name
                for cls_name, cls in sv.all_classes(imports=self.mergeimports).items()
                if not cls.mixin and not cls.abstract
            ]

        if self.split_workbook_by_class:
            output_path = Path(self.schema.name + "_worksheets") if not self.output else Path(self.output)
            output_path = output_path.absolute()

            if not output_path.is_dir():
                output_path.mkdir(parents=True, exist_ok=True)

            for cls_name in classes_to_process:
                cls_output_path = output_path.joinpath(f"{cls_name}.xlsx")
                self.create_workbook_and_worksheets(cls_output_path, [cls_name])
                self.logger.info(f"The Excel workbook for class '{cls_name}' has been written to {cls_output_path}")
        else:
            output_path = Path(self.schema.name + ".xlsx") if not self.output else Path(self.output)
            output_path = output_path.absolute()

            self.create_workbook_and_worksheets(output_path, classes_to_process)
            self.logger.info(f"The Excel workbook has been written to {output_path}")


@shared_arguments(ExcelGenerator)
@click.command(name="excel")
@click.option(
    "--split-workbook-by-class",
    is_flag=True,
    default=False,
    help="""Split model into separate Excel workbooks/files, one for each class""",
)
@click.option(
    "--include-mixins",
    is_flag=True,
    default=False,
    help="""Include mixin classes in the generated Excel workbook/workbooks""",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(),
    help="""Name of Excel spreadsheet to be created, or name of directory to create split workbooks in""",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, split_workbook_by_class, include_mixins, **kwargs):
    """Generate Excel representation of a LinkML model"""
    ExcelGenerator(
        yamlfile, split_workbook_by_class=split_workbook_by_class, include_mixins=include_mixins, **kwargs
    ).serialize(**kwargs)


if __name__ == "__main__":
    cli()



from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition
from linkml_runtime.utils.schemaview import SchemaView


def get_type_designator_value(sv: SchemaView, type_designator_slot: SlotDefinition, class_def: ClassDefinition) -> str:
    """
    returns the correct value for a type designator field for a given class, depending on its range
    this implements the logic described in https://github.com/linkml/linkml/issues/945:
    """
    slot_types = set(sv.type_ancestors(type_designator_slot.range))
    if "uri" in slot_types:
        return sv.get_uri(class_def, expand=True, native=False)
    elif "uriorcurie" in slot_types:
        return sv.get_uri(class_def, expand=False, native=False)
    elif "string" in slot_types:
        return class_def.name
    else:
        return sv.get_uri(class_def, expand=False, native=False)


def get_accepted_type_designator_values(
    sv: SchemaView, type_designator_slot: SlotDefinition, class_def: ClassDefinition
) -> list[str]:
    """
    returns the accepted values for a type designator field for a given class, depending on its range
    this implements the logic described in https://github.com/linkml/linkml/issues/945:
    """
    accepted_uri_values = [
        sv.get_uri(class_def, expand=True, native=True),
        sv.get_uri(class_def, expand=True, native=False),
        sv.get_uri(class_def, expand=False, native=True),
        sv.get_uri(class_def, expand=False, native=False),
    ]
    # unique, but with order preserved (https://stackoverflow.com/a/17016257)
    accepted_uri_values = list(dict.fromkeys(accepted_uri_values))

    slot_types = set(sv.type_ancestors(type_designator_slot.range))
    uri_types = ["uri", "uriorcurie"]

    if slot_types.intersection(uri_types):
        return accepted_uri_values
    elif type_designator_slot.range == "string":
        return [class_def.name]
    else:
        return accepted_uri_values



import importlib
import sys
from collections import Counter
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional, Union

import click
import yaml
from pydantic import BaseModel, Field

from linkml._version import __version__
from linkml.utils import datautils
from linkml.validator import Validator
from linkml.validator.loaders import Loader, default_loader_for_file
from linkml.validator.plugins import ValidationPlugin
from linkml.validator.report import Severity


class Config(BaseModel):
    schema_path: Union[str, Path] = Field(alias="schema")
    target_class: Optional[str] = None
    data_sources: Iterable[Union[str, dict[str, dict[str, str]]]] = []
    plugins: Optional[dict[str, Optional[dict[str, Any]]]] = {"JsonschemaValidationPlugin": {"closed": True}}


def _resolve_class(full_class_name: str, default_package: str, **kwargs):
    if "." in full_class_name:
        module_name, class_name = full_class_name.rsplit(".", maxsplit=1)
    else:
        module_name = default_package
        class_name = full_class_name
    try:
        module = importlib.import_module(module_name)
        class_inst = getattr(module, class_name)
    except (ModuleNotFoundError, AttributeError):
        raise click.ClickException(f"Unknown class: {full_class_name}")
    return class_inst(**kwargs)


def _resolve_plugins(plugin_config: dict[str, dict[str, Any]]) -> list[ValidationPlugin]:
    plugins = []
    for key, value in plugin_config.items():
        plugin = _resolve_class(key, "linkml.validator.plugins", **value if value else {})
        plugins.append(plugin)
    return plugins


def _resolve_loaders(loader_config: Iterable[Union[str, dict[str, dict[str, str]]]]) -> list[Loader]:
    loaders = []
    for entry in loader_config:
        if isinstance(entry, str):
            loader = default_loader_for_file(entry)
        elif isinstance(entry, dict):
            if len(entry) > 1:
                raise click.ClickException(f"Invalid config. Dictionary entries should only have one key: {entry}")
            class_name = next(iter(entry.keys()))
            loader = _resolve_class(class_name, "linkml.validator.loaders", **entry[class_name])
        else:
            raise click.ClickException("Invalid config. Data sources must be specified as a string or as dictionary.")
        loaders.append(loader)
    return loaders


DEPRECATED = "[DEPRECATED: only used in legacy mode]"


@click.command(name="validate")
@click.option(
    "-s",
    "--schema",
    help="Schema file to validate data against",
    type=click.Path(exists=True, dir_okay=False, file_okay=True, resolve_path=True, path_type=Path),
)
@click.option(
    "-C",
    "--target-class",
    help="Class within the schema to validate data against",
)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Validation configuration YAML file.",
)
@click.option(
    "--exit-on-first-failure",
    is_flag=True,
    default=False,
    help="Exit after the first validation failure is found. If not specified all validation failures are reported.",
)
@click.option(
    "--legacy-mode",
    is_flag=True,
    default=False,
    help="Use legacy linkml-validate behavior.",
)
@click.option("--module", "-m", help=f"{DEPRECATED} Path to python datamodel module")
@click.option(
    "--input-format",
    "-f",
    type=click.Choice(list(datautils.dumpers_loaders.keys())),
    help=f"{DEPRECATED} Input format. Inferred from input suffix if not specified",
)
@click.option("--index-slot", "-S", help=f"{DEPRECATED} top level slot. Required for CSV dumping/loading")
@click.option(
    "--include-range-class-descendants/--no-range-class-descendants",
    default=False,
    show_default=False,
    help=f"{DEPRECATED} When handling range constraints, include all descendants of the range "
    "class instead of just the range class",
)
@click.option(
    "--include-context/--no-include-context",
    "-D",
    default=False,
    show_default=True,
    help="Include additional context when reporting of validation errors.",
)
@click.argument("data_sources", nargs=-1, type=click.Path(exists=True))
@click.version_option(__version__, "-V", "--version")
@click.pass_context
def cli(
    context: click.Context,
    schema: Optional[Path],
    target_class: Optional[str],
    config: Optional[str],
    data_sources: tuple[str],
    exit_on_first_failure: bool,
    legacy_mode: bool,
    module: Optional[str],
    input_format: Optional[str],
    index_slot: Optional[str],
    include_range_class_descendants: bool,
    include_context: bool,
):
    """
    Validate data according to a LinkML Schema
    """
    if legacy_mode:
        from linkml.validators import jsonschemavalidator

        return context.invoke(
            jsonschemavalidator.cli,
            input=data_sources[0] if len(data_sources) > 0 else None,
            module=module,
            target_class=target_class,
            input_format=input_format,
            schema=str(schema) if schema else None,
            index_slot=index_slot,
            exit_on_first_failure=exit_on_first_failure,
            include_range_class_descendants=include_range_class_descendants,
        )

    if module is not None:
        click.secho("Warning: the -m/--module option is deprecated except in legacy mode.", fg="yellow")
    if input_format is not None:
        click.secho(
            "Warning: the -f/--input-format option is deprecated except in legacy mode. By "
            "default source loaders are automatically chosen based on file extension. If a loader "
            "needs to manually specified use the data_sources section of a custom config file.",
            fg="yellow",
        )
    if index_slot is not None:
        click.secho("Warning: the -S/--index-slot option is deprecated except in legacy mode.", fg="yellow")
    if include_range_class_descendants:
        click.secho(
            "Warning: the --include-range-class-descendants deprecated except in legacy mode. "
            "This behavior is now the default of the JsonschemaValidationPlugin. If you need to "
            "disable it use the plugins section of a custom config file.",
            fg="yellow",
        )

    config_args = {
        "schema": schema,
        "target_class": target_class,
        "data_sources": data_sources,
    }
    if config:
        with open(config) as config_file:
            config_raw = yaml.safe_load(config_file)
            config_args.update(config_raw)

    if config_args.get("schema") is None:
        raise click.ClickException(
            "No schema specified. Path to schema must be specified by either "
            "the -s/--schema option or in a config file."
        )
    config = Config(**config_args)

    plugins = _resolve_plugins(config.plugins) if config.plugins else []
    loaders = _resolve_loaders(config.data_sources)
    validator = Validator(config.schema_path, validation_plugins=plugins, strict=exit_on_first_failure)
    severity_counter = Counter()
    for loader in loaders:
        for result in validator.iter_results_from_source(loader, config.target_class):
            severity_counter[result.severity] += 1
            click.echo(f"[{result.severity.value}] [{loader.source}/{result.instance_index}] {result.message}")
            if include_context:
                for ctx in result.context:
                    click.echo(f"[CONTEXT] {ctx}")

    if sum(severity_counter.values()) == 0:
        click.echo("No issues found")

    exit_code = 1 if severity_counter[Severity.ERROR] > 0 else 0
    sys.exit(exit_code)


if __name__ == "__main__":
    cli()



"""
Generators translate between a SchemaDefinition and an alternative
representation such as JsonSchema
"""

from linkml.generators.javagen import JavaGenerator
from linkml.generators.jsonldcontextgen import ContextGenerator
from linkml.generators.jsonldgen import JSONLDGenerator
from linkml.generators.jsonschemagen import JsonSchemaGenerator
from linkml.generators.owlgen import OwlSchemaGenerator
from linkml.generators.panderagen import PanderaGenerator
from linkml.generators.pydanticgen import PydanticGenerator
from linkml.generators.pythongen import PythonGenerator
from linkml.generators.rdfgen import RDFGenerator
from linkml.generators.rustgen import RustGenerator
from linkml.generators.shaclgen import ShaclGenerator
from linkml.generators.shexgen import ShExGenerator
from linkml.generators.sqlalchemygen import SQLAlchemyGenerator
from linkml.generators.sqltablegen import SQLTableGenerator

__all__ = [
    "csvgen",
    "dotgen",
    "docgen",
    "golrgen",
    "graphqlgen",
    "javagen",
    "jsonldcontextgen",
    "jsonldgen",
    "jsonschemagen",
    "markdowngen",
    "namespacegen",
    "owlgen",
    "plantumlgen",
    "protogen",
    "pythongen",
    "pydanticgen",
    "rdfgen",
    "shexgen",
    "shaclgen",
    "sssomgen",
    "summarygen",
    "yamlgen",
    "yumlgen",
    "OwlSchemaGenerator",
    "PydanticGenerator",
    "PanderaGenerator",
    "PythonGenerator",
    "JavaGenerator",
    "ContextGenerator",
    "JSONLDGenerator",
    "JsonSchemaGenerator",
    "RustGenerator",
    "ShaclGenerator",
    "ShExGenerator",
    "SQLAlchemyGenerator",
    "SQLTableGenerator",
    "RDFGenerator",
]

# TODO: deprecate usage of these
# GENERATOR_BASE = "0.9"

# PYTHON_GEN_VERSION = GENERATOR_BASE + ".0"



"""
Generate CSVs
"""

import os
from csv import DictWriter
from dataclasses import dataclass
from io import StringIO
from typing import Optional

import click
from linkml_runtime.linkml_model.meta import ClassDefinition, ClassDefinitionName
from linkml_runtime.utils.formatutils import be, underscore

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class CsvGenerator(Generator):
    """
    Generates CSV summaries

    Note: this generator is not widely used,
    and has largely been supplanted by schemasheets
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["csv", "tsv"]
    uses_schemaloader = True
    requires_metamodel = False

    # ObjectVars
    sep: Optional[str] = None
    """Separator for columns"""

    closure: Optional[set[ClassDefinitionName]] = None
    """List of classes to include in output"""

    writer: Optional[DictWriter] = None
    """Python dictwriter"""

    _str_io: Optional[StringIO] = None
    """String that the writer outputs to"""

    def __post_init__(self):
        super().__post_init__()
        self._str_io = None

        self.generate_header()  # TODO: don't do this in initialization

    def generate_header(self) -> str:
        out = f"# metamodel_version: {self.schema.metamodel_version}"
        if self.schema.version:
            out = "\n".join([out, f"# version: {self.schema.version}"])
        return out

    def visit_schema(self, classes: list[ClassDefinitionName] = None, **_) -> None:
        # Note: classes comes from the "root" argument
        self.closure = set()

        if classes is None:
            classes = []

        # Validate the supplied list of classes
        for clsname in classes:
            if clsname not in self.schema.classes:
                raise ValueError(f"Unrecognized class: {clsname}")
            else:
                self.closure.update(self.ancestors(self.schema.classes[clsname]))

        self._str_io = StringIO()
        dialect: str = "excel" if self.format == "csv" else "excel-tab"
        self.writer = DictWriter(self._str_io, ["id", "mappings", "description"], dialect=dialect)
        self.writer.writeheader()

    def visit_class(self, cls: ClassDefinition) -> bool:
        # TODO: find out what to do with mappings
        if not self.closure or cls.name in self.closure:
            self.writer.writerow(
                {
                    "id": underscore(cls.name),
                    # 'mappings': "|".join(cls.mappings),
                    "mappings": "",
                    "description": be(cls.description),
                }
            )
            return True
        return False

    def end_schema(self, **kwargs) -> str:
        return self._str_io.getvalue()


@shared_arguments(CsvGenerator)
@click.command(name="csv")
@click.version_option(__version__, "-V", "--version")
@click.option("--root", "-r", multiple=True, help="Class(es) to transform")
def cli(yamlfile, root=None, **args):
    """Generate CSV/TSV file from LinkML model"""
    print(CsvGenerator(yamlfile, **args).serialize(classes=root, **args))


if __name__ == "__main__":
    cli()



import logging
import os
import pathlib
import sys

import click
import yaml
from linkml_runtime.linkml_model import Prefix
from linkml_runtime.utils import inference_utils
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.inference_utils import infer_all_slot_values
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.generators.pythongen import PythonGenerator
from linkml.utils import datautils, validation
from linkml.utils.datautils import (
    _get_context,
    _get_format,
    _is_xsv,
    dumpers_loaders,
    get_dumper,
    get_loader,
    infer_index_slot,
    infer_root_class,
)

logger = logging.getLogger(__name__)


@click.command(name="convert")
@click.option("--module", "-m", help="Path to python datamodel module")
@click.option("--output", "-o", help="Path to output file")
@click.option(
    "--input-format",
    "-f",
    type=click.Choice(list(dumpers_loaders.keys())),
    help="Input format. Inferred from input suffix if not specified",
)
@click.option(
    "--output-format",
    "-t",
    type=click.Choice(list(dumpers_loaders.keys())),
    help="Output format. Inferred from output suffix if not specified",
)
@click.option(
    "--target-class",
    "-C",
    help="name of class in datamodel that the root node instantiates",
)
@click.option(
    "--target-class-from-path/--no-target-class-from-path",
    default=False,
    show_default=True,
    help="Infer the target class from the filename, should be ClassName-<other-chars>.{yaml,json,...}",
)
@click.option("--index-slot", "-S", help="top level slot. Required for CSV dumping/loading")
@click.option("--schema", "-s", help="Path to schema specified as LinkML yaml")
@click.option("--prefix", "-P", multiple=True, help="Prefixmap base=URI pairs")
@click.option("--prefix-file", help="Path to yaml file containing base=URI pairs")
@click.option(
    "--validate/--no-validate",
    default=True,
    show_default=True,
    help="Validate against the schema",
)
@click.option(
    "--infer/--no-infer",
    default=False,
    show_default=True,
    help="Infer missing slot values",
)
@click.option("--context", "-c", multiple=True, help="path to JSON-LD context file")
@click.version_option(__version__, "-V", "--version")
@click.argument("input")
def cli(
    input,
    module,
    target_class,
    context=None,
    output=None,
    input_format=None,
    output_format=None,
    prefix=None,
    prefix_file=None,
    target_class_from_path=None,
    schema=None,
    validate=None,
    infer=None,
    index_slot=None,
) -> None:
    """
    Converts instance data to and from different LinkML Runtime serialization formats.

    The instance data must conform to a LinkML model, and either a path to a python
    module must be passed, or a path to a schema.

    The converter works by first using a linkml-runtime *loader* to
    instantiate in-memory model objects, then a *dumper* is used to serialize.
    A validation step is optionally performed in between

    When converting to or from RDF, a path to a schema must be provided.

    For more information, see https://linkml.io/linkml/data/index.html
    """
    if prefix and prefix_file is not None:
        raise Exception("Either set prefix OR prefix_file, not both.")
    if prefix is None:
        prefix = []
    if module is None:
        if schema is None:
            raise Exception("must pass one of module OR schema")
        else:
            python_module = PythonGenerator(schema).compile_module()
    else:
        python_module = compile_python(module)
    prefix_map = {}
    if prefix:
        for p in prefix:
            base, uri = p.split("=")
            prefix_map[base] = uri
    if prefix_file is not None:
        prefix_path = pathlib.Path(prefix_file).resolve()
        if not prefix_path.exists():
            raise Exception(f"Path {prefix_file} to prefix map does not exists.")
        with open(prefix_path) as prefix_stream:
            raw_prefix_map = yaml.safe_load(prefix_stream)
        prefix_file_map = raw_prefix_map.get("prefixes", None)
        if prefix_file_map is None:
            raise Exception("Provided prefix file does not contain the prefixes key.")
        prefix_map = prefix_file_map

    if schema is not None:
        sv = SchemaView(schema)
        if prefix_map:
            for k, v in prefix_map.items():
                sv.schema.prefixes[k] = Prefix(k, v)
                sv.set_modified()
    if target_class is None and target_class_from_path:
        target_class = os.path.basename(input).split("-")[0]
        logger.info(f"inferred target class = {target_class} from {input}")
    if target_class is None:
        target_class = infer_root_class(sv)
    if target_class is None:
        raise Exception("target class not specified and could not be inferred")
    py_target_class = python_module.__dict__[target_class]
    input_format = _get_format(input, input_format)
    loader = get_loader(input_format)

    inargs = {}
    outargs = {}
    if datautils._is_rdf_format(input_format):
        if sv is None:
            raise Exception("Must pass schema arg")
        inargs["schemaview"] = sv
        inargs["fmt"] = input_format
    if _is_xsv(input_format):
        if index_slot is None:
            index_slot = infer_index_slot(sv, target_class)
            if index_slot is None:
                raise Exception("--index-slot is required for CSV input")
        inargs["index_slot"] = index_slot
        inargs["schema"] = schema
    obj = loader.load(source=input, target_class=py_target_class, **inargs)
    if infer:
        infer_config = inference_utils.Config(use_expressions=True, use_string_serialization=True)
        infer_all_slot_values(obj, schemaview=sv, config=infer_config)
    if validate:
        if schema is None:
            raise Exception("--schema must be passed in order to validate. Suppress with --no-validate")
        # TODO: use validator framework
        validation.validate_object(obj, schema)

    output_format = _get_format(output, output_format, default="json")
    if output_format == "json-ld":
        if len(context) == 0:
            if schema is not None:
                context = [_get_context(schema)]
            else:
                raise Exception("Must pass in context OR schema for RDF output")
        outargs["contexts"] = list(context)
    if output_format == "rdf" or output_format == "ttl":
        if sv is None:
            raise Exception("Must pass schema arg")
        outargs["schemaview"] = sv
    if _is_xsv(output_format):
        if index_slot is None:
            index_slot = infer_index_slot(sv, target_class)
            if index_slot is None:
                raise Exception("--index-slot is required for CSV output")
        outargs["index_slot"] = index_slot
        outargs["schema"] = schema
    dumper = get_dumper(output_format)
    if output is not None:
        dumper.dump(obj, output, **outargs)
    else:
        print(dumper.dumps(obj, **outargs))


if __name__ == "__main__":
    cli(sys.argv[1:])



from pathlib import Path
from typing import Optional, get_args

import click

from linkml._version import __version__
from linkml.generators.rustgen import RUST_MODES, RustGenerator
from linkml.utils.generator import shared_arguments


@shared_arguments(RustGenerator)
@click.option(
    "-m",
    "--mode",
    type=click.Choice([a for a in get_args(RUST_MODES)]),
    default="crate",
    help="Generation mode: 'crate' (Cargo package) or 'file' (single .rs)",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Overwrite output if it already exists",
)
@click.option(
    "-p",
    "--pyo3",
    is_flag=True,
    help=(
        "Add 'pyo3' to Cargo.toml default features and emit Python module glue (cdylib + #[pymodule]). "
        'Source always includes #[cfg(feature="pyo3")] gates; this flag only enables the crate feature by default.'
    ),
)
@click.option(
    "-s",
    "--serde",
    is_flag=True,
    help=(
        "Add 'serde' to Cargo.toml default features. Source always includes #[cfg(feature=\"serde\")] derives/attrs; "
        "this flag only enables the crate feature by default."
    ),
)
@click.option(
    "--handwritten-lib/--no-handwritten-lib",
    default=False,
    help=(
        "When enabled, place generated sources under src/generated and create a shim lib.rs for handwritten code. "
        "The shim is only created on first run and left untouched on subsequent regenerations."
    ),
)
@click.option("-n", "--crate-name", type=str, default=None, help="Name of the generated crate/module")
@click.option(
    "-o",
    "--output",
    type=click.Path(dir_okay=True),
    help="Output directory (crate mode) or .rs file (file mode)",
)
@click.version_option(__version__, "-V", "--version")
@click.command(name="rust")
def cli(
    yamlfile: Path,
    mode: RUST_MODES = "crate",
    force: bool = False,
    pyo3: bool = False,
    serde: bool = False,
    crate_name: Optional[str] = None,
    handwritten_lib: bool = False,
    output: Optional[Path] = None,
    **kwargs,
):
    gen = RustGenerator(
        yamlfile,
        mode=mode,
        pyo3=pyo3,
        serde=serde,
        output=output,
        crate_name=crate_name,
        handwritten_lib=handwritten_lib,
        **kwargs,
    )
    serialized = gen.serialize(force=force)
    if output is None:
        print(serialized)



from linkml_runtime.linkml_model import ClassDefinition, EnumDefinitionName, SlotDefinition
from rdflib import Literal, URIRef

from linkml.generators.common.ifabsent_processor import IfAbsentProcessor
from linkml.generators.shacl.shacl_data_type import ShaclDataType


class ShaclIfAbsentProcessor(IfAbsentProcessor):
    def map_custom_default_values(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition) -> (bool, str):
        if default_value == "class_curie":
            class_uri = self.schema_view.get_uri(cls, expand=True)
            if class_uri:
                return True, URIRef(class_uri)
            return True, ""

        return False, None

    def map_enum_default_value(
        self, enum_name: EnumDefinitionName, permissible_value_name: str, slot: SlotDefinition, cls: ClassDefinition
    ):
        return Literal(permissible_value_name)

    def map_string_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(default_value, datatype=ShaclDataType.STRING.uri_ref)

    def map_integer_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(default_value, datatype=ShaclDataType.INTEGER.uri_ref)

    def map_boolean_true_default_value(self, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(True, datatype=ShaclDataType.BOOLEAN.uri_ref)

    def map_boolean_false_default_value(self, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(False, datatype=ShaclDataType.BOOLEAN.uri_ref)

    def map_float_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(default_value, datatype=ShaclDataType.FLOAT.uri_ref)

    def map_double_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(default_value, datatype=ShaclDataType.DOUBLE.uri_ref)

    def map_decimal_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(default_value, datatype=ShaclDataType.DECIMAL.uri_ref)

    def map_time_default_value(self, hour: str, minutes: str, seconds: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(f"{hour}:{minutes}:{seconds}", datatype=ShaclDataType.TIME.uri_ref)

    def map_date_default_value(self, year: str, month: str, day: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(f"{year}-{month}-{day}", datatype=ShaclDataType.DATE.uri_ref)

    def map_datetime_default_value(
        self,
        year: str,
        month: str,
        day: str,
        hour: str,
        minutes: str,
        seconds: str,
        slot: SlotDefinition,
        cls: ClassDefinition,
    ):
        return Literal(f"{year}-{month}-{day}T{hour}:{minutes}:{seconds}", datatype=ShaclDataType.DATETIME.uri_ref)

    def map_uri_or_curie_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        uri = URIRef(self.schema_view.expand_curie(default_value))
        return Literal(uri, datatype=ShaclDataType.URI.uri_ref)

    def map_curie_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(default_value, datatype=ShaclDataType.CURIE.uri_ref)

    def map_uri_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return Literal(default_value, datatype=ShaclDataType.URI.uri_ref)

    def map_nc_name_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_object_identifier_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_node_identifier_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_json_pointer_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_json_path_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_sparql_path_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()



import json
import logging
import os
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Optional, Union

import click
from linkml_runtime.linkml_model.meta import (
    AnonymousClassExpression,
    AnonymousSlotExpression,
    ClassDefinition,
    ClassDefinitionName,
    EnumDefinition,
    PermissibleValue,
    PermissibleValueText,
    PresenceEnum,
    SlotDefinition,
    metamodel_version,
)
from linkml_runtime.utils.formatutils import be, camelcase, underscore

from linkml._version import __version__
from linkml.generators.common import build
from linkml.generators.common.lifecycle import LifecycleMixin
from linkml.generators.common.type_designators import get_type_designator_value
from linkml.utils.generator import Generator, shared_arguments
from linkml.utils.helpers import get_range_associated_slots

logger = logging.getLogger(__name__)

# Map from underlying python data type to json equivalent
# Note: The underlying types are a union of any built-in python datatype + any type defined in
#       linkml-runtime/utils/metamodelcore.py
# Note the keys are all lower case
json_schema_types: dict[str, tuple[str, Optional[str]]] = {
    "int": ("integer", None),
    "integer": ("integer", None),
    "bool": ("boolean", None),
    "boolean": ("boolean", None),
    "float": ("number", None),
    "double": ("number", None),
    "decimal": ("number", None),
    "xsddate": ("string", "date"),
    "xsddatetime": ("string", "date-time"),
    "xsdtime": ("string", "time"),
}


class JsonSchema(dict):
    OPTIONAL_IDENTIFIER_SUFFIX = "__identifier_optional"
    PRESERVE_NAMES: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lax_forward_refs = {}

    def add_def(self, name: str, subschema: "JsonSchema") -> None:
        canonical_name = name if self.PRESERVE_NAMES else camelcase(name)

        if "$defs" not in self:
            self["$defs"] = {}

        if "title" not in subschema:
            subschema["title"] = canonical_name

        self["$defs"][canonical_name] = subschema

        if canonical_name in self._lax_forward_refs:
            identifier_name = self._lax_forward_refs.pop(canonical_name)
            self.add_lax_def(canonical_name, identifier_name)

    def add_lax_def(self, names: Union[str, list[str]], identifier_name: str) -> None:
        # JSON-Schema does not have inheritance,
        # so we duplicate slots from inherited parents and mixins
        # Maps e.g. Person --> Person__identifier_optional
        # for use when Person is a range of an inlined-as-dict slot
        if isinstance(names, str):
            names = [names]

        for name in names:
            canonical_name = name if self.PRESERVE_NAMES else camelcase(name)

            if "$defs" not in self or canonical_name not in self["$defs"]:
                self._lax_forward_refs[canonical_name] = identifier_name
            else:
                lax_cls = deepcopy(self["$defs"][canonical_name])
                lax_cls["required"].remove(identifier_name)
                self["$defs"][canonical_name + self.OPTIONAL_IDENTIFIER_SUFFIX] = lax_cls

    def add_property(
        self, name: str, subschema: "JsonSchema", *, value_required: bool = False, value_disallowed: bool = False
    ) -> None:
        canonical_name = name if self.PRESERVE_NAMES else underscore(name)

        if "properties" not in self:
            self["properties"] = {}

        self["properties"][canonical_name] = subschema

        if value_required:
            if "required" not in self:
                self["required"] = []

            self["required"].append(canonical_name)

        # JSON Schema does not have a very natural way to express that a property cannot be present.
        # The apparent best way to do it is to use:
        # {
        #   properties: {
        #     foo: ...
        #   },
        #   not: {
        #     required: ['foo']
        #   }
        # }
        # The {required: [foo]} subschema evaluates to true if the foo property is present with any
        # value. Wrapping that in a `not` keyword inverts that condition.
        if value_disallowed:
            if "not" not in self:
                self["not"] = {}
            if "required" not in self["not"]:
                self["not"]["required"] = []

            self["not"]["required"].append(canonical_name)

    def add_keyword(self, keyword: str, value: Any):
        if value is None:
            return

        self[keyword] = value

    @property
    def is_array(self):
        typ = self.get("type", False)
        if isinstance(typ, str):
            return typ == "array"
        elif isinstance(typ, list):
            return "array" in typ
        else:
            return False

    @property
    def is_object(self):
        return self.get("type") == "object"

    def to_json(self, **kwargs) -> str:
        return json.dumps(self, **kwargs)

    @classmethod
    def ref_for(cls, class_name: Union[str, list[str]], identifier_optional: bool = False, required: bool = True):
        def _ref(class_name):
            def_name = class_name if cls.PRESERVE_NAMES else camelcase(class_name)
            def_suffix = cls.OPTIONAL_IDENTIFIER_SUFFIX if identifier_optional else ""
            return JsonSchema({"$ref": f"#/$defs/{def_name}{def_suffix}"})

        if isinstance(class_name, list):
            if len(class_name) == 1:
                ref = _ref(class_name[0])
            else:
                ref = JsonSchema({"anyOf": [_ref(name) for name in class_name]})
        else:
            ref = _ref(class_name)

        if not required:
            if "anyOf" in ref:
                ref["anyOf"].append({"type": "null"})
            else:
                ref = JsonSchema({"anyOf": [ref, {"type": "null"}]})
        return ref

    @classmethod
    def array_of(cls, subschema: "JsonSchema", include_null: bool, required: bool = True) -> "JsonSchema":
        if required or not include_null:
            typ = "array"
        else:
            typ = ["array", "null"]

        schema = {"type": typ, "items": subschema}

        return JsonSchema(schema)


class SchemaResult(build.SchemaResult):
    """Top-level result of building a json schema"""

    schema_: JsonSchema


class EnumResult(build.EnumResult):
    """A single built enum"""

    schema_: JsonSchema


class ClassResult(build.ClassResult):
    """A single built class"""

    schema_: JsonSchema


class SlotResult(build.SlotResult):
    """A slot within the context of a class"""

    schema_: JsonSchema


@dataclass
class JsonSchemaGenerator(Generator, LifecycleMixin):
    """
    Generates JSONSchema documents from a LinkML SchemaDefinition

    - Each linkml class generates a schema
    - inheritance hierarchies are rolled-down from ancestors
    - Composition not yet implemented
    - Enumerations treated as strings
    - Foreign key references are treated as semantics-free strings

    This generator implements the following :class:`.LifecycleMixin` methods:

    * :meth:`.LifecycleMixin.before_generate_schema`
    * :meth:`.LifecycleMixin.after_generate_schema`
    * :meth:`.LifecycleMixin.before_generate_classes`
    * :meth:`.LifecycleMixin.before_generate_enums`
    * :meth:`.LifecycleMixin.before_generate_class_slots`
    * :meth:`.LifecycleMixin.before_generate_class`
    * :meth:`.LifecycleMixin.after_generate_class`
    * :meth:`.LifecycleMixin.before_generate_class_slot`
    * :meth:`.LifecycleMixin.after_generate_class_slot`
    * :meth:`.LifecycleMixin.before_generate_enum`
    * :meth:`.LifecycleMixin.after_generate_enum`

    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.3"
    valid_formats = ["json"]
    uses_schemaloader = False
    file_extension = "schema.json"
    materialize_patterns: bool = False

    # @deprecated("Use top_class")
    topClass: Optional[str] = None

    not_closed: Optional[bool] = True
    """If not closed, then an open-ended set of attributes can be instantiated for any object"""

    indent: int = 4

    inline: bool = False

    top_class: Optional[Union[ClassDefinitionName, str]] = None  # JSON object is one instance of this
    """Class instantiated by the root node of the document tree"""

    include_range_class_descendants: bool = False
    """If set, use an open world assumption and allow the range of a slot to be any descendant of the declared range.
    Note that if the range of a slot has a type designator, descendants will always be included.
    """

    title_from: str = "name"
    """The slot from which to populate JSONSchema title annotation."""

    top_level_schema: JsonSchema = None

    include_null: bool = True
    """Whether to include a "null" type in optional slots"""

    preserve_names: bool = False
    """If true, preserve LinkML element names in JSON Schema output (e.g., for $defs, properties, $ref targets)."""

    def __post_init__(self):
        if self.topClass:
            logger.warning("topClass is deprecated - use top_class")
            self.top_class = self.topClass

        super().__post_init__()

        # Set the class variable for JsonSchema to use
        JsonSchema.PRESERVE_NAMES = self.preserve_names

        if self.top_class:
            if self.schemaview.get_class(self.top_class) is None:
                logger.warning(f"No class in schema named {self.top_class}")

    def start_schema(self, inline: bool = False):
        self.inline = inline

        self.top_level_schema = JsonSchema(
            {
                "$schema": "https://json-schema.org/draft/2019-09/schema",
                "$id": self.schema.id,
                "metamodel_version": metamodel_version,
                "version": self.schema.version if self.schema.version else None,
                "title": self.schema.title if self.title_from == "title" and self.schema.title else self.schema.name,
                "type": "object",
                "additionalProperties": self.not_closed,
            }
        )

    def handle_class(self, cls: ClassDefinition) -> None:
        cls = self.before_generate_class(cls, self.schemaview)

        if cls.mixin or cls.abstract:
            return

        subschema_type = "object"
        additional_properties = False
        if self.is_class_unconstrained(cls):
            subschema_type = ["null", "boolean", "object", "number", "string"]
            additional_properties = True

        class_subschema = JsonSchema(
            {
                "type": subschema_type,
                "additionalProperties": additional_properties,
                "description": be(cls.description),
            }
        )
        if self.title_from == "title" and cls.title:
            class_subschema["title"] = cls.title

        class_slots = self.before_generate_class_slots(
            self.schemaview.class_induced_slots(cls.name), cls, self.schemaview
        )
        for slot_definition in class_slots:
            self.handle_class_slot(subschema=class_subschema, cls=cls, slot=slot_definition)

        rule_subschemas = []
        for ancestor_class_name in self.schemaview.class_ancestors(cls.name):
            ancestor_class = self.schemaview.get_class(ancestor_class_name)
            for rule in ancestor_class.rules:
                subschema = JsonSchema()

                open_world = rule.open_world
                if open_world is None:
                    open_world = False

                if_subschema = self.get_subschema_for_anonymous_class(rule.preconditions, properties_required=True)
                if if_subschema:
                    subschema["if"] = if_subschema

                then_subschema = self.get_subschema_for_anonymous_class(
                    rule.postconditions, properties_required=not open_world
                )
                if then_subschema:
                    subschema["then"] = then_subschema

                # same as required requirements as postconditions?
                else_subschema = self.get_subschema_for_anonymous_class(
                    rule.elseconditions, properties_required=not open_world
                )
                if else_subschema:
                    subschema["else"] = else_subschema

                rule_subschemas.append(subschema)

                if rule.bidirectional:
                    inverse_subschema = JsonSchema()

                    if then_subschema:
                        inverse_subschema["if"] = then_subschema

                    if if_subschema:
                        inverse_subschema["then"] = if_subschema

                    rule_subschemas.append(inverse_subschema)

        if len(rule_subschemas) == 1:
            class_subschema.update(rule_subschemas[0])
        elif len(rule_subschemas) > 1:
            if "allOf" not in class_subschema:
                class_subschema["allOf"] = []
            class_subschema["allOf"].extend(rule_subschemas)

        class_subschema = self.after_generate_class(
            ClassResult.model_construct(schema_=class_subschema, source=cls), self.schemaview
        ).schema_

        self.top_level_schema.add_def(cls.name, class_subschema)

        if (
            self.top_class is not None
            and (
                (self.preserve_names and self.top_class == cls.name)
                or (not self.preserve_names and camelcase(self.top_class) == camelcase(cls.name))
            )
        ) or (self.top_class is None and cls.tree_root):
            for key, value in class_subschema.items():
                # check this first to ensure we don't overwrite things like additionalProperties
                # or description on the root. But we do want to copy over properties, required,
                # if, then, etc.
                if key not in self.top_level_schema:
                    self.top_level_schema[key] = value

    def get_subschema_for_anonymous_class(
        self, cls: AnonymousClassExpression, properties_required: bool = False
    ) -> Union[None, JsonSchema]:
        if not cls:
            return None

        subschema = JsonSchema()
        for slot in cls.slot_conditions.values():
            prop = self.get_subschema_for_slot(slot, omit_type=True, include_null=False)
            value_required = False
            value_disallowed = False
            if slot.value_presence:
                if slot.value_presence == PresenceEnum(PresenceEnum.PRESENT):
                    value_required = True
                elif slot.value_presence == PresenceEnum(PresenceEnum.ABSENT):
                    value_disallowed = True
            elif slot.required is not None:
                value_required = slot.required
            else:
                value_required = properties_required
            subschema.add_property(
                self.aliased_slot_name(slot), prop, value_required=value_required, value_disallowed=value_disallowed
            )

        if cls.any_of is not None and len(cls.any_of) > 0:
            subschema["anyOf"] = [self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.any_of]

        if cls.all_of is not None and len(cls.all_of) > 0:
            subschema["allOf"] = [self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.all_of]

        if cls.exactly_one_of is not None and len(cls.exactly_one_of) > 0:
            subschema["oneOf"] = [
                self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.exactly_one_of
            ]

        if cls.none_of is not None and len(cls.none_of) > 0:
            subschema["not"] = {
                "anyOf": [self.get_subschema_for_anonymous_class(c, properties_required) for c in cls.any_of]
            }

        return subschema

    def handle_enum(self, enum: EnumDefinition) -> None:
        # TODO: this only works with explicitly permitted values. It will need to be extended to
        # support other pv_formula
        enum = self.before_generate_enum(enum, self.schemaview)

        def extract_permissible_text(pv):
            if isinstance(pv, str):
                return pv
            if isinstance(pv, PermissibleValue):
                return pv.text.code
            if isinstance(pv, PermissibleValueText):
                return pv
            raise ValueError(f"Invalid permissible value in enum {enum}: {pv}")

        permissible_values_texts = list(map(extract_permissible_text, enum.permissible_values or []))

        enum_schema = JsonSchema(
            {
                "type": "string",
                "description": be(enum.description),
            }
        )
        if self.title_from == "title" and enum.title:
            enum_schema["title"] = enum.title

        if permissible_values_texts:
            enum_schema["enum"] = permissible_values_texts

        enum_schema = self.after_generate_enum(
            EnumResult.model_construct(schema_=enum_schema, source=enum), self.schemaview
        ).schema_
        self.top_level_schema.add_def(enum.name, enum_schema)

    def get_type_info_for_slot_subschema(
        self, slot: Union[SlotDefinition, AnonymousSlotExpression]
    ) -> tuple[str, str, Union[str, list[str]]]:
        # JSON Schema type (https://json-schema.org/understanding-json-schema/reference/type.html)
        typ = None
        # Reference to a JSON schema entity (https://json-schema.org/understanding-json-schema/structuring.html#ref)
        reference = None
        # JSON Schema format (https://json-schema.org/understanding-json-schema/reference/string.html#format)
        fmt = None

        slot_is_inlined = self.schemaview.is_inlined(slot)
        if slot.range in self.schemaview.all_types().keys():
            schema_type = self.schemaview.induced_type(slot.range)
            (typ, fmt) = json_schema_types.get(schema_type.base.lower(), ("string", None))
        elif slot.range in self.schemaview.all_enums().keys():
            reference = slot.range
        elif slot.range in self.schemaview.all_classes().keys():
            if slot_is_inlined:
                descendants = [
                    desc
                    for desc in self.schemaview.class_descendants(slot.range)
                    if not self.schemaview.get_class(desc).abstract
                ]
                # Always include class descendants if the range class has a type designator
                include_range_class_descendants = (
                    self.include_range_class_descendants
                    or self.schemaview.get_type_designator_slot(slot.range) is not None
                )
                if descendants and include_range_class_descendants:
                    reference = descendants
                else:
                    reference = slot.range
            else:
                id_slot = self.schemaview.get_identifier_slot(slot.range)
                return self.get_type_info_for_slot_subschema(id_slot)

        return (typ, fmt, reference)

    def get_value_constraints_for_slot(self, slot: Union[SlotDefinition, AnonymousSlotExpression, None]) -> JsonSchema:
        if slot is None:
            return JsonSchema()

        constraints = JsonSchema()
        if slot.range in self.schemaview.all_types().keys():
            # types take lower priority
            schema_type = self.schemaview.induced_type(slot.range)
            constraints.add_keyword("pattern", schema_type.pattern)
            constraints.add_keyword("minimum", schema_type.minimum_value)
            constraints.add_keyword("maximum", schema_type.maximum_value)
            constraints.add_keyword("const", schema_type.equals_string)
            constraints.add_keyword("const", schema_type.equals_number)
        constraints.add_keyword("pattern", slot.pattern)
        constraints.add_keyword("minimum", slot.minimum_value)
        constraints.add_keyword("maximum", slot.maximum_value)
        constraints.add_keyword("const", slot.equals_string)
        constraints.add_keyword("const", slot.equals_number)
        if slot.equals_string_in:
            constraints.add_keyword("enum", slot.equals_string_in)
        return constraints

    def get_subschema_for_slot(
        self, slot: Union[SlotDefinition, AnonymousSlotExpression], omit_type: bool = False, include_null: bool = True
    ) -> JsonSchema:
        """
        Args:
            include_null: Include ``type: null`` when generating ranges that are not required
        """
        prop = JsonSchema()
        if isinstance(slot, SlotDefinition) and slot.array:
            # TODO: this is currently too lax, in that it will validate ANY array.
            # see https://github.com/linkml/linkml/issues/2188
            prop = JsonSchema(
                {
                    "type": ["null", "boolean", "object", "number", "string", "array"],
                    "additionalProperties": True,
                }
            )
            return JsonSchema.array_of(prop, include_null, required=slot.required)
        slot_is_multivalued = "multivalued" in slot and slot.multivalued
        slot_is_inlined = self.schemaview.is_inlined(slot)
        slot_is_boolean = any([slot.any_of, slot.all_of, slot.exactly_one_of, slot.none_of])
        if not omit_type:
            typ, fmt, reference = self.get_type_info_for_slot_subschema(slot)
            if slot_is_inlined:
                # If inline we have to include redefined slots
                if slot_is_multivalued:
                    (
                        range_id_slot,
                        range_simple_dict_value_slot,
                        range_required_slots,
                    ) = get_range_associated_slots(self.schemaview, slot.range)
                    # if the range class has an ID and the slot is not inlined as a list, then we need to consider
                    # various inlined as dict formats
                    if range_id_slot is not None and not slot.inlined_as_list:
                        # At a minimum, the inlined dict can have keys (additionalProps) that are IDs
                        # and the values are the range class but possibly omitting the ID.
                        additionalProps = [JsonSchema.ref_for(reference, identifier_optional=True)]

                        # If the range can be collected as a simple dict, then we can also accept the value
                        # of that simple dict directly.
                        if range_simple_dict_value_slot is not None:
                            additionalProps.append(
                                self.get_subschema_for_slot(range_simple_dict_value_slot, include_null=False)
                            )

                        # If the range has no required slots, then null is acceptable
                        if len(range_required_slots) == 0:
                            additionalProps.append(JsonSchema({"type": "null"}))

                        # If through the above logic we identified multiple acceptable forms, then wrap them
                        # in an "anyOf", otherwise just take the only acceptable form
                        if len(additionalProps) == 1:
                            additionalProps = additionalProps[0]
                        else:
                            additionalProps = JsonSchema({"anyOf": additionalProps})
                        if slot.required or not include_null:
                            typ = "object"
                        else:
                            typ = ["object", "null"]
                        prop = JsonSchema({"type": typ, "additionalProperties": additionalProps})
                        self.top_level_schema.add_lax_def(reference, self.aliased_slot_name(range_id_slot))
                    else:
                        prop = JsonSchema.array_of(JsonSchema.ref_for(reference), include_null, required=slot.required)
                else:
                    prop = JsonSchema.ref_for(reference, required=slot.required or not include_null)

            else:
                if reference is not None:
                    prop = JsonSchema.ref_for(reference)
                elif typ and fmt is None:
                    prop = JsonSchema({"type": typ})
                elif typ:
                    prop = JsonSchema({"type": typ, "format": fmt})

                if slot_is_multivalued:
                    prop = JsonSchema.array_of(prop, include_null, required=slot.required)
                else:
                    # handle optionals - bools like any_of, etc. below as they call this method recursively
                    if not slot.required and not slot_is_boolean and include_null:
                        if "type" in prop:
                            prop["type"] = [prop["type"], "null"]

        prop.add_keyword("description", slot.description)
        if self.title_from == "title" and slot.title:
            prop.add_keyword("title", slot.title)

        own_constraints = self.get_value_constraints_for_slot(slot)

        if prop.is_array:
            all_element_constraints = self.get_value_constraints_for_slot(slot.all_members)
            any_element_constraints = self.get_value_constraints_for_slot(slot.has_member)
            prop.add_keyword("minItems", slot.minimum_cardinality)
            prop.add_keyword("maxItems", slot.maximum_cardinality)
            prop["items"].update(own_constraints)
            prop["items"].update(all_element_constraints)
            if any_element_constraints:
                prop["contains"] = any_element_constraints
        else:
            prop.update(own_constraints)

        if prop.is_object:
            prop.add_keyword("minProperties", slot.minimum_cardinality)
            prop.add_keyword("maxProperties", slot.maximum_cardinality)

        bool_subschema = JsonSchema()
        if slot.any_of is not None and len(slot.any_of) > 0:
            bool_subschema["anyOf"] = [self.get_subschema_for_slot(s, include_null=False) for s in slot.any_of]
            if not slot.required and not prop.is_array and include_null:
                bool_subschema["anyOf"].append({"type": "null"})

        if slot.all_of is not None and len(slot.all_of) > 0:
            bool_subschema["allOf"] = [self.get_subschema_for_slot(s, include_null=False) for s in slot.all_of]

        if slot.exactly_one_of is not None and len(slot.exactly_one_of) > 0:
            bool_subschema["oneOf"] = [self.get_subschema_for_slot(s, include_null=False) for s in slot.exactly_one_of]

        if slot.none_of is not None and len(slot.none_of) > 0:
            bool_subschema["not"] = {
                "anyOf": [self.get_subschema_for_slot(s, include_null=False) for s in slot.none_of]
            }

        if bool_subschema:
            if prop.is_array:
                if "items" not in prop:
                    prop["items"] = {}
                if slot.required or not include_null:
                    prop["type"] = "array"
                else:
                    prop["type"] = ["array", "null"]
                prop["items"].update(bool_subschema)
            else:
                prop.update(bool_subschema)

        return prop

    def handle_class_slot(self, subschema: JsonSchema, cls: ClassDefinition, slot: SlotDefinition) -> None:
        slot = self.before_generate_class_slot(slot, cls, self.schemaview)
        class_id_slot = self.schemaview.get_identifier_slot(cls.name, use_key=True)
        value_required = (
            slot.required or slot == class_id_slot or slot.value_presence == PresenceEnum(PresenceEnum.PRESENT)
        )
        value_disallowed = slot.value_presence == PresenceEnum(PresenceEnum.ABSENT)

        aliased_slot_name = self.aliased_slot_name(slot)
        prop = self.get_subschema_for_slot(slot, include_null=self.include_null)
        prop = self.after_generate_class_slot(
            SlotResult.model_construct(schema_=prop, source=slot), cls, self.schemaview
        ).schema_
        subschema.add_property(
            aliased_slot_name, prop, value_required=value_required, value_disallowed=value_disallowed
        )

        if slot.designates_type:
            type_value = get_type_designator_value(self.schemaview, slot, cls)
            prop["enum"] = [type_value]

    def generate(self) -> JsonSchema:
        self.schema = self.before_generate_schema(self.schema, self.schemaview)
        self.start_schema()

        all_enums = self.before_generate_enums(self.schemaview.all_enums().values(), self.schemaview)
        for enum_definition in all_enums:
            self.handle_enum(enum_definition)

        all_classes = self.before_generate_classes(self.schemaview.all_classes().values(), self.schemaview)
        for class_definition in all_classes:
            self.handle_class(class_definition)

        self.top_level_schema = self.after_generate_schema(
            SchemaResult.model_construct(schema_=self.top_level_schema, source=self.schema), self.schemaview
        ).schema_
        return self.top_level_schema

    def serialize(self, **kwargs) -> str:
        if self.materialize_patterns:
            logger.info("Materializing patterns in the schema before serialization")
            self.schemaview.materialize_patterns()
        result = self.generate().to_json(sort_keys=True, indent=self.indent if self.indent > 0 else None)
        return result.rstrip() + "\n"


@shared_arguments(JsonSchemaGenerator)
@click.command(name="json-schema")
@click.option(
    "-i",
    "--inline",
    is_flag=True,
    help="""
Generate references to types rather than inlining them.
Note that declaring a slot as inlined: true will always inline the class
""",
)
@click.option(
    "-t",
    "--top-class",
    help="""
Top level class; slots of this class will become top level properties in the json-schema
""",
)
@click.option(
    "--not-closed/--closed",
    default=True,
    show_default=True,
    help="""
Set additionalProperties=False if closed otherwise true if not closed at the global level
""",
)
@click.option(
    "--include-range-class-descendants/--no-range-class-descendants",
    default=False,
    show_default=False,
    help="""
When handling range constraints, include all descendants of the range class instead of just the range class
""",
)
@click.option(
    "--indent",
    default=4,
    show_default=True,
    help="""
If this is a positive number the resulting JSON will be pretty-printed with that indent level. Set to 0 to
disable pretty-printing and return the most compact JSON representation
""",
)
@click.option(
    "--title-from",
    type=click.Choice(["name", "title"], case_sensitive=False),
    default="name",
    help="""
Specify from which slot are JSON Schema 'title' annotations generated.
""",
)
@click.option(
    "-d",
    "--include",
    help="""
Include LinkML Schema outside of imports mechanism.  Helpful in including deprecated classes and slots in a separate
YAML, and including it when necessary but not by default (e.g. in documentation or for backwards compatibility)
""",
)
@click.option(
    "--materialize-patterns/--no-materialize-patterns",
    default=True,  # Default set to True
    show_default=True,
    help="If set, patterns will be materialized in the generated JSON Schema.",
)
@click.option(
    "--preserve-names/--normalize-names",
    default=False,
    show_default=True,
    help="Preserve original LinkML names in JSON Schema output (e.g., for $defs, properties, $ref targets).",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **kwargs):
    """Generate JSON Schema representation of a LinkML model"""
    print(JsonSchemaGenerator(yamlfile, **kwargs).serialize(**kwargs))


if __name__ == "__main__":
    cli()



"""Generate plantuml

https://plantuml.com/

"""

import base64
import os
import zlib
from dataclasses import dataclass
from typing import Callable, Optional, cast

import click
import requests
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    SlotDefinition,
)
from linkml_runtime.utils.formatutils import camelcase, underscore

from linkml import REQUESTS_TIMEOUT
from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

plantuml_is_a = "^--"
plantuml_uses = ("--", " : uses")
plantuml_injected = ("--", " : inject")
plantuml_inline = "*--"
plantuml_inline_rev = "--*"
plantuml_ref = "--"


@dataclass
class PlantumlGenerator(Generator):
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["puml", "plantuml", "png", "pdf", "jpg", "json", "svg"]
    visit_all_class_slots = False
    preserve_names: bool = False

    referenced: Optional[set[ClassDefinitionName]] = None  # List of classes that have to be emitted
    generated: Optional[set[ClassDefinitionName]] = None  # List of classes that have been emitted
    class_generated: Optional[set[ClassDefinitionName]] = None  # Class definitions that have been emitted
    associations_generated: Optional[set[ClassDefinitionName]] = None  # Classes with associations generated
    focus_classes: Optional[set[ClassDefinitionName]] = None  # Classes to be completely filled
    gen_classes: Optional[set[ClassDefinitionName]] = None  # Classes to be generated
    output_file_name: Optional[str] = None  # Location of output file if directory used

    classes: set[ClassDefinitionName] = None
    directory: Optional[str] = None
    kroki_server: Optional[str] = "https://kroki.io"
    tooltips_flag: bool = False
    dry_run: bool = False

    def visit_schema(
        self,
        classes: set[ClassDefinitionName] = None,
        directory: Optional[str] = None,
        **_,
    ) -> Optional[str]:
        if directory:
            os.makedirs(directory, exist_ok=True)
        if classes is not None:
            for cls in classes:
                if cls not in self.schema.classes:
                    raise ValueError(f"Unknown class name: {cls}")
        self.class_generated = set()
        self.associations_generated = set()
        if classes is None:
            self.focus_classes = set()
        else:
            self.focus_classes = set(classes)
        if classes:
            self.gen_classes = self.neighborhood(list(classes)).classrefs.union(classes)
        else:
            self.gen_classes = self.synopsis.roots.classrefs
        self.referenced = self.gen_classes
        self.generated = set()
        plantumlclassdef: list[str] = []
        while self.referenced.difference(self.generated):
            cn = sorted(list(self.referenced.difference(self.generated)), reverse=True)[0]
            self.generated.add(cn)
            assocs = self.class_associations(ClassDefinitionName(cn), cn in self.referenced)
            if assocs:
                plantumlclassdef.extend(assocs)
            else:
                if ClassDefinitionName(cn) not in self.class_generated:
                    plantumlclassdef.append(self.add_class(ClassDefinitionName(cn)))
                    self.add_class(ClassDefinitionName(cn))

        dedup_plantumlclassdef = []
        [dedup_plantumlclassdef.append(x) for x in plantumlclassdef if x not in dedup_plantumlclassdef]

        plantuml_code = "\n".join(dedup_plantumlclassdef)
        b64_diagram = base64.urlsafe_b64encode(zlib.compress(plantuml_code.encode(), 9))

        plantuml_url = self.kroki_server + "/plantuml/svg/" + b64_diagram.decode()
        if self.dry_run:
            return plantuml_url
        if directory:
            file_suffix = ".svg" if self.format == "puml" or self.format == "puml" else "." + self.format
            schema_name = sorted(classes)[0] if classes else self.schema.name
            filename = schema_name if self.preserve_names else camelcase(schema_name)
            self.output_file_name = os.path.join(directory, filename + file_suffix)
            resp = requests.get(plantuml_url, stream=True, timeout=REQUESTS_TIMEOUT)
            if resp.ok:
                with open(self.output_file_name, "wb") as f:
                    for chunk in resp.iter_content(chunk_size=2048):
                        f.write(chunk)
            else:
                self.logger.error(f"{resp.reason} accessing {plantuml_url}")
        else:
            out = (
                "@startuml\n"
                "skinparam nodesep 10\n"
                "hide circle\n"
                "hide empty members\n" + "\n".join(dedup_plantumlclassdef) + "\n@enduml\n"
            )
            return out

    def add_class(self, cn: ClassDefinitionName) -> str:
        """Define the class only if
        (a) it hasn't previously been generated and
        (b) it appears in the gen_classes list

        @param cn:
        @return:
        """
        slot_defs: list[str] = []
        if cn not in self.class_generated and (not self.focus_classes or cn in self.focus_classes):
            cls = self.schema.classes[cn]
            for slot in self.filtered_cls_slots(cn, all_slots=True, filtr=lambda s: s.range not in self.schema.classes):
                if True or cn in slot.domain_of:
                    mod = self.prop_modifier(cls, slot)
                    slot_name = (
                        self.aliased_slot_name(slot)
                        if self.preserve_names
                        else underscore(self.aliased_slot_name(slot))
                    )
                    range_name = slot.range if self.preserve_names else underscore(slot.range)
                    slot_defs.append(
                        "    {field} " + slot_name + mod + " : " + range_name + " " + self.cardinality(slot)
                    )
            self.class_generated.add(cn)
        self.referenced.add(cn)
        cls = self.schema.classes[cn]

        tooltip_contents = str(cls.description)
        first_newline_index = tooltip_contents.find("\n")
        tooltip_contents = tooltip_contents if first_newline_index < 0 else tooltip_contents[0:first_newline_index]

        if self.format == "svg" and len(tooltip_contents) > 200:
            tooltip_contents = tooltip_contents[0:197] + " ... "

        tooltip = " [[{" + tooltip_contents + "}]] "
        if cls.abstract:
            class_type = "abstract"
        else:
            class_type = "class"
        return class_type + ' "' + cn + '"' + tooltip + ("{\n" + "\n".join(slot_defs) + "\n}")

    def class_associations(self, cn: ClassDefinitionName, must_render: bool = False) -> str:
        """Emit all associations for a focus class.  If none are specified, all classes are generated

        @param cn: Name of class to be emitted
        @param must_render: True means render even if this is a target (class is specifically requested)
        @return: PLANTUML representation of the association
        """

        classes: list[str] = []
        assocs: list[str] = []
        if cn not in self.associations_generated and (not self.focus_classes or cn in self.focus_classes):
            cls = self.schema.classes[cn]

            # Slots that reference other classes
            for slot in self.filtered_cls_slots(cn, False, lambda s: s.range in self.schema.classes)[::-1]:
                # Swap the two boxes because, in the case of self reference, the last definition wins
                if slot.range not in self.associations_generated and cn in slot.domain_of:
                    if cn not in self.class_generated:
                        classes.append(self.add_class(cn))
                    lhs = cn
                    if cast(ClassDefinitionName, slot.range) not in self.class_generated:
                        classes.append(self.add_class(cast(ClassDefinitionName, slot.range)))
                    rhs = cast(ClassDefinitionName, slot.range)
                    assocs.append(
                        '"'
                        + lhs
                        + '" '
                        + (plantuml_inline if slot.inlined else plantuml_ref)
                        + "> "
                        + self.cardinality(slot, False)
                        + '"'
                        + rhs
                        + '" : '
                        + '"'
                        + self.aliased_slot_name(slot)
                        + '"'
                        + self.prop_modifier(cls, slot)
                    )

            # Slots in other classes that reference this
            for slotname in sorted(self.synopsis.rangerefs.get(cn, [])):
                slot = self.schema.slots[slotname]
                # Don't do self references twice
                # Also, slot must be owned by the class
                if cls.name not in slot.domain_of and cls.name not in self.associations_generated:
                    for dom in [self.schema.classes[dof] for dof in slot.domain_of]:
                        if dom.name not in self.class_generated:
                            classes.append(self.add_class(dom.name))
                        if cn not in self.class_generated:
                            classes.append(self.add_class(cn))
                        assocs.append(
                            '"'
                            + dom.name
                            + '" '
                            + (plantuml_inline if slot.inlined else plantuml_ref)
                            + "> "
                            + self.cardinality(slot, False)
                            + '"'
                            + cn
                            + '" : '
                            + '"'
                            + self.aliased_slot_name(slot)
                            + '"'
                            + self.prop_modifier(dom, slot)
                        )

            # Mixins used in the class
            for mixin in cls.mixins:
                if cn not in self.class_generated:
                    classes.append(self.add_class(cn))
                if mixin not in self.class_generated:
                    classes.append(self.add_class(mixin))
                assocs.append('"' + cn + '" ' + plantuml_uses[0] + ' "' + mixin + '" ' + plantuml_uses[1])

            # Classes that use the class as a mixin
            if cls.name in self.synopsis.mixinrefs:
                for mixin in sorted(self.synopsis.mixinrefs[cls.name].classrefs, reverse=True):
                    if ClassDefinitionName(mixin) not in self.class_generated:
                        classes.append(self.add_class(ClassDefinitionName(mixin)))
                    if cn not in self.class_generated:
                        classes.append(self.add_class(cn))
                    assocs.append(
                        '"' + ClassDefinitionName(mixin) + '" ' + plantuml_uses[0] + ' "' + cn + '" ' + plantuml_uses[1]
                    )

            # Classes that inject information
            if cn in self.synopsis.applytos.classrefs:
                for injector in sorted(self.synopsis.applytorefs[cn].classrefs, reverse=True):
                    if cn not in self.class_generated:
                        classes.append(self.add_class(cn))
                    if ClassDefinitionName(injector) not in self.class_generated:
                        classes.append(self.add_class(ClassDefinitionName(injector)))
                    assocs.append(cn + plantuml_injected[0] + ClassDefinitionName(injector) + plantuml_injected[1])
            self.associations_generated.add(cn)

            # Children
            if cn in self.synopsis.isarefs:
                for is_a_cls in sorted(self.synopsis.isarefs[cn].classrefs, reverse=True):
                    if cn not in self.class_generated:
                        classes.append(self.add_class(cn))
                    if ClassDefinitionName(is_a_cls) not in self.class_generated:
                        classes.append(self.add_class(ClassDefinitionName(is_a_cls)))
                    assocs.append('"' + cn + '" ' + plantuml_is_a + ' "' + ClassDefinitionName(is_a_cls) + '"')

            # Parent
            if cls.is_a and cls.is_a not in self.associations_generated:
                if cls.is_a not in self.class_generated:
                    classes.append(self.add_class(cls.is_a))
                if cn not in self.class_generated:
                    classes.append(self.add_class(cn))
                assocs.append('"' + cls.is_a + '" ' + plantuml_is_a + ' "' + cn + '"')
        entries: list[str] = []
        entries.extend(classes)
        entries.extend(assocs)
        return entries

    @staticmethod
    def cardinality(slot: SlotDefinition, is_attribute: bool = True) -> str:
        if is_attribute:
            if slot.multivalued:
                return " [1..*]" if slot.required else " [0..*]"
            else:
                return " "
        else:
            if slot.multivalued:
                return '"1..*" ' if slot.required else '"0..*" '
            else:
                return '"1" ' if slot.required else '"0..1" '

    def filtered_cls_slots(
        self,
        cn: ClassDefinitionName,
        all_slots: bool = True,
        filtr: Callable[[SlotDefinition], bool] = None,
    ) -> list[SlotDefinition]:
        """Return the set of slots associated with the class that meet the filter criteria.  Slots will be returned
        in defining order, with class slots returned last

        @param cn: name of class to filter
        @param all_slots: True means include attributes
        @param filtr: Slot filter predicate
        @return: List of slot definitions
        """
        if filtr is None:

            def filtr(x):
                return True

        rval = []
        cls = self.schema.classes[cn]
        cls_slots = self.all_slots(cls, cls_slots_first=True)
        for slot in cls_slots:
            if (all_slots or slot.range in self.schema.classes) and filtr(slot):
                rval.append(slot)

        return rval

    def prop_modifier(self, cls: ClassDefinition, slot: SlotDefinition) -> str:
        """Return the modifiers for the slot:
            (i) - inherited
            (m) - inherited through mixin
            (a) - injected
            (pk) - primary ckey

        @param cls:
        @param slot:
        @return:
        """
        pk = "(pk)" if slot.key else ""
        inherited = slot.name not in self.own_slot_names(cls)
        mixin = inherited and slot.name in [mslot.name for mslot in [self.schema.classes[m] for m in cls.mixins]]
        injected = cls.name in self.synopsis.applytos.classrefs and slot.name in [
            aslot.name
            for aslot in [self.schema.classes[a] for a in sorted(self.synopsis.applytorefs[cls.name].classrefs)]
        ]
        return pk + ("(a)" if injected else "(m)" if mixin else "(i)" if inherited else "")


@shared_arguments(PlantumlGenerator)
@click.command(name="plantuml")
@click.option("--classes", "-c", multiple=True, help="Class(es) to emit")
@click.option(
    "--directory",
    "-d",
    help="Output directory - if supplied, PlantUML rendering will be saved in file",
)
@click.option(
    "--kroki-server",
    "-k",
    help="URL of the Kroki server to use for diagram drawing",
    default="https://kroki.io",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    show_default=True,
    help="Print out Kroki URL calls instead of sending the real requests",
)
@click.option(
    "--preserve-names/--normalize-names",
    default=False,
    show_default=True,
    help="Preserve original LinkML names in PlantUML diagram output (e.g., for class names, slot names, file names).",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **args):
    """Generate a UML representation of a LinkML model"""
    print(PlantumlGenerator(yamlfile, **args).serialize(**args), end="")


if __name__ == "__main__":
    cli()



from collections.abc import Iterator
from typing import Any

from linkml_runtime.loaders import json_loader

from linkml.validator.loaders.loader import Loader


class JsonLoader(Loader):
    """A loader for instances serialized as JSON

    :param source: Path or URL to JSON source
    """

    def __init__(self, source) -> None:
        super().__init__(source)

    def iter_instances(self) -> Iterator[Any]:
        """Lazily yield instance from JSON source.

        If the root of the JSON is an array, yield each element of the array. Otherwise,
        yield the root element itself.

        :return: Iterator over data instances
        :rtype: Iterator[Any]
        """
        data = json_loader.load_as_dict(self.source)
        if isinstance(data, list):
            yield from data
        else:
            yield data
